import twaddle.lexer.lexer as lexer
from twaddle.lexer.lexer_tokens import Token, TokenType


def test_angle_brackets():
    test_string = "<>"
    result = lexer.lex(test_string)

    expected_result = [
        Token(TokenType.LEFT_ANGLE_BRACKET),
        Token(TokenType.RIGHT_ANGLE_BRACKET),
    ]

    assert len(result) == len(expected_result)
    for actual, expected in zip(result, expected_result):
        assert actual == expected


def test_square_brackets():
    test_string = "[]"
    result = lexer.lex(test_string)

    expected_result = [
        Token(TokenType.LEFT_SQUARE_BRACKET),
        Token(TokenType.RIGHT_SQUARE_BRACKET),
    ]

    assert len(result) == len(expected_result)
    for actual, expected in zip(result, expected_result):
        assert actual == expected


def test_curly_brackets():
    test_string = "{}"
    result = lexer.lex(test_string)

    expected_result = [
        Token(TokenType.LEFT_CURLY_BRACKET),
        Token(TokenType.RIGHT_CURLY_BRACKET),
    ]

    assert len(result) == len(expected_result)
    for actual, expected in zip(result, expected_result):
        assert actual == expected


def test_pipe():
    test_string = "|"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.PIPE)

    assert len(result) == 1
    assert result[0] == expected_result


def test_hyphen():
    test_string = "-"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.HYPHEN)

    assert len(result) == 1
    assert result[0] == expected_result


def test_semicolon():
    test_string = ";"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.SEMICOLON)

    assert len(result) == 1
    assert result[0] == expected_result


def test_colon():
    test_string = ":"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.COLON)

    assert len(result) == 1
    assert result[0] == expected_result


def test_double_colon():
    test_string = "::"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.DOUBLE_COLON)

    assert len(result) == 1
    assert result[0] == expected_result


def test_quote():
    test_string = '"'
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.QUOTE)

    assert len(result) == 1
    assert result[0] == expected_result


def test_new_line():
    test_string = r"\n"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.NEW_LINE)

    assert len(result) == 1
    assert result[0] == expected_result


def test_escaped_space():
    test_string = chr(92) + "s"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.SPACE)

    assert len(result) == 1
    assert result[0] == expected_result


def test_indefinite_article():
    test_string = r"\a"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.LOWER_INDEFINITE_ARTICLE)

    assert len(result) == 1
    assert result[0] == expected_result


def test_slash():
    test_string = r"\\"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.DOUBLE_BACKSLASH)

    assert len(result) == 1
    assert result[0] == expected_result


def test_digit():
    test_string = r"\d"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.DIGIT)

    assert len(result) == 1
    assert result[0] == expected_result


def test_plaintext():
    test_string = "hello"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.PLAIN_TEXT, "hello")

    assert len(result) == 1
    assert result[0] == expected_result


def test_exclamation_mark():
    test_string = "!"
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.EXCLAMATION_MARK)

    assert len(result) == 1
    assert result[0] == expected_result


def test_equals():
    test_string = "="
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.EQUALS)

    assert len(result) == 1
    assert result[0] == expected_result


def test_dot():
    test_string = "."
    result = lexer.lex(test_string)

    expected_result = Token(TokenType.DOT)

    assert len(result) == 1
    assert result[0] == expected_result


def test_long_string():
    test_string = (
        "angle brackets <:> curly brackets{ ::  } square brackets []plaintext-hello|"
    )
    result = lexer.lex(test_string)

    expected_result = [
        Token(TokenType.PLAIN_TEXT, "angle brackets "),
        Token(TokenType.LEFT_ANGLE_BRACKET),
        Token(TokenType.COLON),
        Token(TokenType.RIGHT_ANGLE_BRACKET),
        Token(TokenType.PLAIN_TEXT, " curly brackets"),
        Token(TokenType.LEFT_CURLY_BRACKET),
        Token(TokenType.PLAIN_TEXT, " "),
        Token(TokenType.DOUBLE_COLON),
        Token(TokenType.PLAIN_TEXT, "  "),
        Token(TokenType.RIGHT_CURLY_BRACKET),
        Token(TokenType.PLAIN_TEXT, " square brackets "),
        Token(TokenType.LEFT_SQUARE_BRACKET),
        Token(TokenType.RIGHT_SQUARE_BRACKET),
        Token(TokenType.PLAIN_TEXT, "plaintext"),
        Token(TokenType.HYPHEN),
        Token(TokenType.PLAIN_TEXT, "hello"),
        Token(TokenType.PIPE),
    ]

    assert len(result) == len(expected_result)
    for actual, expected in zip(result, expected_result):
        assert actual == expected


def test_escaped_characters():
    test_string = r"\<\>\{\}\[\]\|"

    expected_result = [
        Token(TokenType.SINGLE_BACKSLASH),
        Token(TokenType.LEFT_ANGLE_BRACKET),
        Token(TokenType.SINGLE_BACKSLASH),
        Token(TokenType.RIGHT_ANGLE_BRACKET),
        Token(TokenType.SINGLE_BACKSLASH),
        Token(TokenType.LEFT_CURLY_BRACKET),
        Token(TokenType.SINGLE_BACKSLASH),
        Token(TokenType.RIGHT_CURLY_BRACKET),
        Token(TokenType.SINGLE_BACKSLASH),
        Token(TokenType.LEFT_SQUARE_BRACKET),
        Token(TokenType.SINGLE_BACKSLASH),
        Token(TokenType.RIGHT_SQUARE_BRACKET),
        Token(TokenType.SINGLE_BACKSLASH),
        Token(TokenType.PIPE),
    ]
    result = lexer.lex(test_string)
    assert len(result) == 14
    for actual, expected in zip(result, expected_result):
        assert actual == expected


def test_article_in_separator():
    test_string = r"[sep:\a]"
    result = lexer.lex(test_string)

    expected_result = [
        Token(TokenType.LEFT_SQUARE_BRACKET),
        Token(TokenType.PLAIN_TEXT, "sep"),
        Token(TokenType.COLON),
        Token(TokenType.LOWER_INDEFINITE_ARTICLE),
        Token(TokenType.RIGHT_SQUARE_BRACKET),
    ]

    for actual, expected in zip(result, expected_result):
        print(f"{str(actual)=}, {str(expected)=}")
        assert actual == expected


def test_realistic_sentence():
    test_string = r"I work as \a <noun-job>"
    result = lexer.lex(test_string)

    expected_result = [
        Token(TokenType.PLAIN_TEXT, "I work as "),
        Token(TokenType.LOWER_INDEFINITE_ARTICLE),
        Token(TokenType.PLAIN_TEXT, " "),
        Token(TokenType.LEFT_ANGLE_BRACKET),
        Token(TokenType.PLAIN_TEXT, "noun"),
        Token(TokenType.HYPHEN),
        Token(TokenType.PLAIN_TEXT, "job"),
        Token(TokenType.RIGHT_ANGLE_BRACKET),
    ]

    assert len(result) == len(expected_result)
    for actual, expected in zip(result, expected_result):
        assert actual == expected



from copy import copy
from functools import singledispatchmethod
from random import randint, randrange
from re import Match, sub
from typing import Optional

from twaddle.compiler.compiler import Compiler
from twaddle.compiler.compiler_objects import (
    BlockObject,
    DigitObject,
    FunctionObject,
    IndefiniteArticleObject,
    LookupObject,
    RegexObject,
    RootObject,
    TextObject,
)
from twaddle.exceptions import TwaddleInterpreterException
from twaddle.interpreter.block_attributes import BlockAttributeManager, BlockAttributes
from twaddle.interpreter.formatter import Formatter
from twaddle.interpreter.function_dict import function_definitions
from twaddle.interpreter.regex_state import RegexState
from twaddle.interpreter.synchronizer import Synchronizer, SynchronizerManager
from twaddle.lookup.lookup_dictionary import LookupDictionary
from twaddle.lookup.lookup_manager import LookupManager


class Interpreter:
    SPECIAL_FUNCTIONS = ["clear", "load", "paste"]

    def __init__(
        self,
        lookup_manager: LookupManager,
        persistent_labels: bool = False,
        persistent_synchronizers: bool = False,
        persistent_patterns: bool = False,
        persistent_clipboard: bool = False,
        strict_mode: bool = False,
    ):
        self.persistent_labels = persistent_labels
        self.persistent_synchronizers = persistent_synchronizers
        self.persistent_patterns = persistent_patterns
        self.persistent_clipboard = persistent_clipboard
        self.lookup_manager = lookup_manager
        self.synchronizer_manager = SynchronizerManager()
        self.block_attribute_manager = BlockAttributeManager()
        self.compiler = Compiler(strict_mode=strict_mode)
        self.saved_patterns = dict()
        self.copied_blocks = dict()
        self.strict_mode = strict_mode

    def interpret_external(self, sentence: str) -> str:
        self.clear()
        compiled_sentence = self.compiler.compile(sentence)
        return self.interpret_internal(compiled_sentence)

    def interpret_internal(self, parse_result: RootObject) -> str:
        formatter = Formatter()
        for obj in parse_result:
            resulting_formatter = self.run(obj)
            if resulting_formatter:
                formatter += resulting_formatter
        result = formatter.resolve()
        return result

    def clear(self):
        if not self.persistent_labels:
            self.lookup_manager.clear_labels()
        if not self.persistent_synchronizers:
            self.synchronizer_manager.clear()
        if not self.persistent_patterns:
            self.saved_patterns.clear()
        if not self.persistent_clipboard:
            self.copied_blocks.clear()
        self.block_attribute_manager.clear()

    def force_clear(self):
        self.lookup_manager.clear_labels()
        self.synchronizer_manager.clear()
        self.saved_patterns.clear()
        self.block_attribute_manager.clear()
        self.copied_blocks.clear()

    def _get_synchronizer_for_block(
        self, attributes: BlockAttributes, num_choices: int
    ) -> Optional[Synchronizer]:
        if attributes.synchronizer is None:
            return None
        if self.synchronizer_manager.synchronizer_exists(attributes.synchronizer):
            synchronizer = self.synchronizer_manager.get_synchronizer(
                attributes.synchronizer
            )
            if self.strict_mode and synchronizer.num_choices != num_choices:
                raise TwaddleInterpreterException(
                    f"[Interpreter._get_synchronizer_for_block] Invalid number of choices ({num_choices}) "
                    f"for synchronizer '{attributes.synchronizer}', initialised with {synchronizer.num_choices}"
                )
            return synchronizer
        if attributes.synchronizer_type is None:
            raise TwaddleInterpreterException(
                "[Interpreter.run](RantBlockObject) tried to define new synchronizer "
                "without defining synchronizer type"
            )
        return self.synchronizer_manager.create_synchronizer(
            attributes.synchronizer,
            attributes.synchronizer_type,
            num_choices,
        )

    # noinspection PyUnusedLocal
    @singledispatchmethod
    def run(self, arg) -> Formatter:
        formatter = Formatter()
        return formatter

    @run.register(RootObject)
    def _(self, root: RootObject):
        formatter = Formatter()
        for item in root.contents:
            item_result = self.run(item)
            if item_result:
                formatter += item_result
        return formatter

    @run.register(BlockObject)
    def _(self, block: BlockObject):
        formatter = Formatter()
        attributes: BlockAttributes = self.block_attribute_manager.get_attributes()
        first_repetition = True
        if attributes.repetitions < 2:
            attributes.separator = None
            attributes.first = None
            attributes.last = None
        synchronizer = self._get_synchronizer_for_block(attributes, len(block.choices))

        while attributes.repetitions:
            if synchronizer is None:
                choice = randrange(0, len(block.choices))
            else:
                choice = synchronizer.next()
                if choice >= len(block.choices):
                    raise TwaddleInterpreterException(
                        f"[Interpreter.run](RantBlockObject) tried to get item no. {choice} of {len(block.choices)} -"
                        "when using synchronizers, make sure you have the same number of choices each time"
                    )
            if first_repetition and attributes.first:
                first_repetition = False
                formatter.append_formatter(self.run(attributes.first))
            elif attributes.repetitions == 1:
                if attributes.last:
                    formatter.append_formatter(self.run(attributes.last))
                elif attributes.separator:
                    formatter.append_formatter(self.run(attributes.separator))
            attributes.repetitions = attributes.repetitions - 1
            partial_result = self.run(block.choices[choice])
            formatter += partial_result
            if attributes.repetitions > 1 and attributes.separator:
                formatter.append_formatter(self.run(attributes.separator))
        if name := attributes.save_as:
            self.saved_patterns[name] = block
        if name := attributes.copy_as:
            self.copied_blocks[name] = copy(formatter)
        if attributes.hidden:
            return Formatter()
        if attributes.reverse:
            block_result = formatter.resolve()
            formatter = Formatter()
            formatter.append("".join(reversed(block_result)))
        if attributes.abbreviate:
            abbreviation = self._get_abbreviation(formatter, attributes)
            formatter = Formatter()
            formatter.append(abbreviation)
        return formatter

    def _get_abbreviation(
        self, input_formatter: Formatter, attributes: BlockAttributes
    ) -> str:
        output_formatter = Formatter()
        if case := attributes.abbreviation_case:
            output_formatter.set_strategy(case)

        def _get_abbr_component(word: str) -> str:
            for index, char in enumerate(word):
                if char.isalpha():
                    return char
                if char.isdigit():
                    digits = ""
                    while char.isdigit():
                        digits += str(char)
                        index += 1
                        if index >= len(word):
                            break
                        char = word[index]
                    return digits
            return ""

        abbreviation = "".join(
            [_get_abbr_component(word) for word in input_formatter.resolve().split()]
        )
        output_formatter.append(abbreviation)
        return output_formatter.resolve()

    def _handle_special_functions(
        self, func: FunctionObject, evaluated_args: list[str]
    ):
        match func.func:
            case "clear":
                self.force_clear()
                return Formatter()
            case "load":
                if not len(evaluated_args):
                    raise TwaddleInterpreterException(
                        "[Interpreter._handle_special_functions] Tried "
                        "to load pattern without specifying name"
                    )
                if not (block := self.saved_patterns.get(evaluated_args[0])):
                    raise TwaddleInterpreterException(
                        "[Interpreter._handle_special_functions#load] Tried "
                        f"to load unknown pattern '{evaluated_args[0]}'"
                    )
                return self.run(block)
            case "paste":
                if not len(evaluated_args):
                    raise TwaddleInterpreterException(
                        "[Interpreter._handle_special_functions] Tried "
                        "to paste block result without specifying name"
                    )
                if not (block := self.copied_blocks.get(evaluated_args[0])):
                    raise TwaddleInterpreterException(
                        "[Interpreter._handle_special_functions#paste] Tried "
                        f"to paste result of unknown block '{evaluated_args[0]}'"
                    )
                return block
            case _:
                raise TwaddleInterpreterException(
                    f"[Interpreter] function '{func.func}' "
                    "marked special but no special handling defined"
                )

    def _save_pattern(self, block: BlockObject, name: str):
        self.saved_patterns[name] = block

    @run.register(FunctionObject)
    def _(self, func: FunctionObject):
        formatter = Formatter()
        evaluated_args = list()
        for arg in func.args:
            evaluated_args.append(self.run(arg).resolve())
        if func.func in self.SPECIAL_FUNCTIONS:
            return self._handle_special_functions(func, evaluated_args)
        if func.func in function_definitions:
            formatter.append(
                function_definitions[func.func](
                    evaluated_args, self.block_attribute_manager, func.args
                )
            )
        else:
            raise TwaddleInterpreterException(
                f"[Interpreter::run] no function found named '{func.func}'"
            )
        return formatter

    @run.register(TextObject)
    def _(self, text: TextObject):
        formatter = Formatter()
        formatter.append(text.text)
        return formatter

    @run.register(LookupObject)
    def _(self, lookup: LookupObject):
        formatter = Formatter()
        dictionary: LookupDictionary = self.lookup_manager[lookup.dictionary]
        formatter.append(dictionary.get(lookup))
        return formatter

    # noinspection SpellCheckingInspection
    @run.register(IndefiniteArticleObject)
    def _(self, indef: IndefiniteArticleObject):
        formatter = Formatter()
        formatter.add_indefinite_article(indef.default_upper)
        return formatter

    # noinspection PyUnusedLocal
    @run.register(DigitObject)
    def _(self, digit: DigitObject):
        formatter = Formatter()
        formatter.append(str(randint(0, 9)))
        return formatter

    @run.register(RegexObject)
    def _(self, regex: RegexObject):
        # noinspection SpellCheckingInspection
        formatter = Formatter()

        def repl(matchobj: Match):
            RegexState.match = matchobj.group()
            return self.run(regex.replacement).resolve()

        formatter.append(sub(regex.regex, repl, self.run(regex.scope).resolve()))
        return formatter



from pathlib import Path

from twaddle.lookup.lookup_dictionary import LookupDictionary


class DictionaryFileParser:
    @staticmethod
    def read_forms(forms_line: str) -> list[str]:
        return forms_line.split()[1:]

    @staticmethod
    def read_name(name_line: str) -> str:
        return name_line.split()[1]

    @staticmethod
    def read_entry(entry_line: str) -> list[str]:
        return entry_line.split("/")

    @staticmethod
    def read_from_path(path: Path) -> LookupDictionary:
        with path.open("r", encoding="utf-8") as input_file:
            name = str()
            forms = list[str]()
            dictionary = None
            classes = set[str]()
            for line in input_file:
                if dictionary is None:
                    if not name:
                        name = DictionaryFileParser._try_read_name(line)
                    if not forms:
                        forms = DictionaryFileParser._try_read_forms(line)
                    if name and forms:
                        dictionary = LookupDictionary(name, forms)
                else:
                    DictionaryFileParser._read_line(line, forms, classes, dictionary)
            return dictionary

    @staticmethod
    def _try_read_name(line: str) -> str:
        name = None
        if line.startswith("#name"):
            name = DictionaryFileParser.read_name(line)
        return name

    @staticmethod
    def _try_read_forms(line: str) -> list[str]:
        forms = None
        if line.startswith("#forms") or line.startswith("#subs"):
            forms = DictionaryFileParser.read_forms(line)
        return forms

    @staticmethod
    def _read_line(
        line: str,
        forms: list[str],
        classes: list[str],
        dictionary: LookupDictionary,
    ):
        line = line.strip()
        if line.startswith("#class add"):
            classes.add(line.split()[-1])
        elif line.startswith("#class remove") and line.split()[-1] in classes:
            classes.remove(line.split()[-1])
        elif line.startswith("> "):
            line = line[2:]
            entry = DictionaryFileParser.read_entry(line)
            if len(entry) == len(forms):
                dictionary.add(entry, set.copy(classes))



import re
from functools import singledispatchmethod
from typing import Self, Type

from twaddle.compiler.compiler_objects import IndefiniteArticleObject
from twaddle.exceptions import TwaddleInterpreterException
from twaddle.interpreter.formatting_object import (
    FormattingObject,
    FormattingStrategy,
    IndefiniteArticle,
    PlainText,
    StrategyChange,
)


class Formatter:
    # regex for finding next alphabetic string
    alphabetic_regex = re.compile(r"[^\W_]+", re.UNICODE)

    def __init__(self):
        self.output_stack = list()
        self.sentence = str()
        self.current_strategy = FormattingStrategy.NONE
        self.indefinite_article_waiting = False

    def _reset_(self):
        self.output_stack = list()
        self.sentence = str()
        self.current_strategy = FormattingStrategy.NONE
        self.indefinite_article_waiting = False

    @singledispatchmethod
    def append(self, arg) -> Self:
        if arg is None:
            formatter = Formatter()
            return formatter
        else:
            raise TwaddleInterpreterException(
                f"[Formatter.append] tried to append unexpected type {type(arg)}"
            )

    @append.register(str)
    def _(self, item: str) -> Self:
        previous = self._get_previous_object_()
        self.output_stack.append(PlainText(previous, item))
        if self.indefinite_article_waiting:
            self._replace_indefinite_articles(item)
        return

    @append.register(FormattingStrategy)
    def _(self, item: FormattingStrategy) -> Self:
        self.output_stack.append(StrategyChange(self._get_previous_object_(), item))

    @append.register(IndefiniteArticleObject)
    def _(self, item: IndefiniteArticleObject) -> Self:
        self.add_indefinite_article(item.default_upper)

    @append.register(IndefiniteArticle)
    def _(self, item: IndefiniteArticle) -> Self:
        self.add_indefinite_article(item.default_upper)

    @append.register(PlainText)
    def _(self, item: PlainText) -> Self:
        previous = self._get_previous_object_()
        self.output_stack.append(PlainText(previous, item.text))
        if self.indefinite_article_waiting:
            self._replace_indefinite_articles(item.text)

    @append.register(FormattingObject)
    def _(self, item: FormattingObject) -> Self:
        item.previous = self._get_previous_object_()
        self.output_stack.append(item)

    def append_formatter(self, formatter: Self):
        self.output_stack += formatter.output_stack

    def _get_previous_object_(self) -> Type[FormattingObject] | None:
        if len(self.output_stack):
            return self.output_stack[-1]
        return None

    def _print_(self, text_object: PlainText):
        self._append_to_sentence_(text_object.text)

    def _default_indefinite_article_(self, _: IndefiniteArticle):
        self._append_to_sentence_("a")

    def _append_to_sentence_(self, text: str):
        match self.current_strategy:
            case FormattingStrategy.NONE:
                self.sentence += text
            case FormattingStrategy.UPPER:
                self.sentence += text.upper()
            case FormattingStrategy.LOWER:
                self.sentence += text.lower()
            case FormattingStrategy.SENTENCE:
                self.sentence += self.apply_sentence_case(text)
            case FormattingStrategy.TITLE:
                self.sentence += self.apply_title_case(text)
            case _:
                raise TwaddleInterpreterException(
                    f"[Formatter.append] no handling defined for {self.current_strategy}"
                )

    def __iadd__(self, other):
        for item in other.output_stack:
            self.append(item)
        return self

    def resolve(self) -> str:
        function_dict = {
            PlainText: self._print_,
            StrategyChange: self._set_strategy_,
            IndefiniteArticle: self._default_indefinite_article_,
        }
        for item in self.output_stack:
            function_dict[type(item)](item)
        result = self.sentence
        self._reset_()
        return result

    def set_strategy(self, strategy: FormattingStrategy):
        self.output_stack.append(StrategyChange(self._get_previous_object_(), strategy))

    def _set_strategy_(self, strategy: StrategyChange):
        self.current_strategy = strategy.strategy

    def add_indefinite_article(self, default_upper=False):
        previous = self._get_previous_object_()
        self.output_stack.append(IndefiniteArticle(previous, default_upper))
        self.indefinite_article_waiting = True

    def _replace_indefinite_articles(self, text: str):
        next_word = self._find_alphabetic_string_(text)
        if next_word is None:
            return
        chosen_article = "an" if self._indefinite_article_use_an_(next_word) else "a"
        self.output_stack = [
            (
                item
                if not isinstance(item, IndefiniteArticle)
                else self._convert_article(item, chosen_article)
            )
            for item in self.output_stack
        ]

    def _convert_article(
        self, article: IndefiniteArticle, chosen_article: str
    ) -> PlainText:
        if article.default_upper:
            chosen_article = chosen_article.capitalize()
        self.indefinite_article_waiting = False
        result = PlainText(article.previous, chosen_article)
        if article.next:
            result.next = article.next
            article.next.previous = result
        return PlainText(article.previous, chosen_article)

    def _find_alphabetic_string_(self, text: str) -> str | None:
        result = self.alphabetic_regex.search(text)
        if result:
            return result[0]
        return None

    # noinspection SpellCheckingInspection
    def _indefinite_article_use_an_(self, next_word: str) -> bool:
        # exclude/include prefixes/words taken directly from old version of rant
        # this can definitely be improved but will do for now
        irregular_a_prefixes = {
            "uni",
            "use",
            "uri",
            "urol",
            "U.",
            "one",
            "uvu",
            "eul",
            "euk",
            "eur",
        }
        irregular_an_prefixes = {"honest", "honor", "hour", "8"}
        irregular_a_words = {"u"}
        irregular_an_words = {"f", "fbi", "fcc", "fda", "x", "l", "m", "n", "s", "h"}
        if any(next_word.startswith(prefix) for prefix in irregular_a_prefixes):
            return False
        if any(next_word == word for word in irregular_a_words):
            return False
        if any(next_word.startswith(prefix) for prefix in irregular_an_prefixes):
            return True
        if any(next_word == word for word in irregular_an_words):
            return True
        return next_word[0].lower() in ["a", "e", "i", "o", "u"]

    def apply_sentence_case(self, text: str) -> str:
        # 1st person pronoun is a special case, slightly awkward to
        # handle especially if result of lookup. short circuit and
        # skip rest of evaluation if "I" is the total text content
        if text == "I":
            return text
        result = str()
        sentence_start = False
        if not self.sentence or self.sentence.rstrip()[-1] in ".!?":
            sentence_start = True
        for char in text:
            if not char.isalnum() and len(result) > 0:
                if result[-2:] == " i":
                    result = result[:-2] + " I"
            if sentence_start:
                result += char.upper()
            else:
                result += char.lower()
            if not char.isspace():
                sentence_start = char in ".!?"
        return result

    def apply_title_case(self, text: str) -> str:
        result = str()
        word_start = False
        if not self.sentence or self.sentence[-1].isspace():
            word_start = True
        for char in text:
            if word_start:
                result += char.upper()
            else:
                result += char.lower()
            word_start = True if char.isspace() else False
        return result



from collections import OrderedDict
from random import choice

from twaddle.compiler.compiler_objects import IndefiniteArticleObject, LookupObject
from twaddle.exceptions import TwaddleLookupException
from twaddle.lookup.lookup_entry import DictionaryEntry


class LookupDictionary:
    special_tokens = {"{a}": IndefiniteArticleObject()}

    def __init__(self, name: str, forms: list[str]):
        self.name = name
        self.forms = forms
        self.entries = list[DictionaryEntry]()
        self.labels = dict[str, DictionaryEntry]()
        self.tags = list[str]()

    def add(self, forms: list[str], tags: set[str] = None):
        if len(forms) != len(self.forms):
            raise TwaddleLookupException(
                "[LookupDictionary.add] wrong number of forms provided"
            )
        entry = DictionaryEntry(OrderedDict(zip(self.forms, forms)), tags)
        if tags:
            for tag in tags:
                if tag not in self.tags:
                    self.tags.append(tag)
        self.entries.append(entry)

    def clear_labels(self):
        self.labels = dict[str, DictionaryEntry]()

    def _get_form(self, form: str) -> str:
        if form is None:
            form = self.forms[0]
        if form not in self.forms:
            raise TwaddleLookupException(
                f"[LookupDictionary.get] dictionary '{self.name}' has no form '{form}'"
            )
        return form

    def _valid_choices_for_strictness_level(
        self,
        valid_choices: list[DictionaryEntry],
        lookup: LookupObject,
    ) -> list[DictionaryEntry]:
        if not valid_choices:
            if lookup.strict_mode:
                raise TwaddleLookupException(
                    "[LookupDictionary._valid_choices_for_strictness_level] "
                    f"no valid choices for strict mode lookup in dictionary '{self.name}'"
                )
            valid_choices = list.copy(self.entries)
        return valid_choices

    def _get_valid_choices(
        self,
        lookup: LookupObject,
    ) -> list[DictionaryEntry]:
        valid_choices = list[DictionaryEntry]()
        for entry in self.entries:
            if entry.has_any_tag_of(lookup.negative_tags):
                continue
            if not lookup.positive_tags or entry.has_all_tags(lookup.positive_tags):
                valid_choices.append(entry)
        valid_choices = self._valid_choices_for_strictness_level(valid_choices, lookup)
        valid_choices = self._prune_valid_choices(valid_choices, lookup.negative_labels)
        valid_choices = self._valid_choices_for_strictness_level(valid_choices, lookup)
        return valid_choices

    def _prune_valid_choices(
        self, valid_choices: list[DictionaryEntry], labels_negative: set[str]
    ) -> list[DictionaryEntry]:
        if labels_negative:
            for label in labels_negative:
                if label in self.labels and self.labels[label] in valid_choices:
                    valid_choices.remove(self.labels[label])
        return valid_choices

    def _get(self, lookup: LookupObject) -> str:
        lookup.form = self._get_form(lookup.form)
        if lookup.positive_label and lookup.positive_label in self.labels:
            return self.labels[lookup.positive_label][lookup.form]

        valid_choices = self._get_valid_choices(lookup)

        chosen_entry = choice(valid_choices)
        if lookup.positive_label:
            self.labels[lookup.positive_label] = chosen_entry
        return chosen_entry[lookup.form]

    def _strict_class_validation(self, lookup: LookupObject):
        all_tags = list()
        if lookup.positive_tags:
            for tag in lookup.positive_tags:
                all_tags.append(tag)
        if lookup.negative_tags:
            for tag in lookup.negative_tags:
                all_tags.append(tag)
        for tag in all_tags:
            if tag not in self.tags:
                raise TwaddleLookupException(
                    f"[LookupDictionary._strict_class_validation] Invalid class '{tag}' requested "
                    f"for dictionary '{self.name}' in strict mode"
                )

    def _strict_label_validation(self, lookup: LookupObject):
        if lookup.negative_labels:
            for label in lookup.negative_labels:
                if label not in self.labels:
                    raise TwaddleLookupException(
                        "[LookupDictionary._strict_label_validation] Requested antimatch of label "
                        f"'{label}', not defined for dictionary '{self.name}'"
                    )

    def _validate_strict_mode(self, lookup: LookupObject):
        if not lookup.strict_mode:
            return
        self._strict_class_validation(lookup)
        self._strict_label_validation(lookup)

    def get(self, lookup: LookupObject) -> str | IndefiniteArticleObject:
        self._validate_strict_mode(lookup)
        result = self._get(lookup)
        if result in self.special_tokens:
            return self.special_tokens[result]
        return result



from collections import deque

from twaddle.compiler.compiler_objects import TextObject
from twaddle.exceptions import TwaddleParserException
from twaddle.lexer.lexer_tokens import Token, TokenType


def to_plain_text_token(raw: Token) -> Token:
    match raw.type:
        case TokenType.PLAIN_TEXT:
            return raw
        case TokenType.EXCLAMATION_MARK:
            return Token(TokenType.PLAIN_TEXT, "!")
        case TokenType.COLON:
            return Token(TokenType.PLAIN_TEXT, ":")
        case TokenType.SEMICOLON:
            return Token(TokenType.PLAIN_TEXT, ";")
        case TokenType.PIPE:
            return Token(TokenType.PLAIN_TEXT, "|")
        case TokenType.HYPHEN:
            return Token(TokenType.PLAIN_TEXT, "-")
        case TokenType.DOUBLE_COLON:
            return Token(TokenType.PLAIN_TEXT, "::")
        case TokenType.DOUBLE_BACKSLASH:
            return Token(TokenType.PLAIN_TEXT, "/")
        case TokenType.DOT:
            return Token(TokenType.PLAIN_TEXT, ".")
        case TokenType.EQUALS:
            return Token(TokenType.PLAIN_TEXT, "=")
        # unrecognised token type
        # should throw an error here once everything's ready
        case _:
            return Token(TokenType.PLAIN_TEXT, raw.value)


def to_plain_text_object(raw: Token) -> TextObject:
    return TextObject(get_text_for_object(raw))


def get_text_for_object(raw: Token) -> str:
    match raw.type:
        case TokenType.LEFT_ANGLE_BRACKET:
            return "<"
        case TokenType.RIGHT_ANGLE_BRACKET:
            return ">"
        case TokenType.LEFT_CURLY_BRACKET:
            return "{"
        case TokenType.RIGHT_CURLY_BRACKET:
            return "}"
        case TokenType.LEFT_SQUARE_BRACKET:
            return "["
        case TokenType.RIGHT_SQUARE_BRACKET:
            return "]"
        case TokenType.PIPE:
            return "|"
        case TokenType.HYPHEN:
            return "-"
        case TokenType.SEMICOLON:
            return ";"
        case TokenType.COLON:
            return ":"
        case TokenType.DOUBLE_COLON:
            return "::"
        case TokenType.QUOTE:
            return '"'
        case TokenType.NEW_LINE:
            return "\n"
        case TokenType.TAB:
            return "\t"
        case TokenType.LOWER_INDEFINITE_ARTICLE:
            return r"\a"
        case TokenType.UPPER_INDEFINITE_ARTICLE:
            return r"\A"
        case TokenType.DOUBLE_BACKSLASH:
            return "\\"
        case TokenType.SINGLE_BACKSLASH:
            return "\\"
        case TokenType.FORWARD_SLASH:
            return "/"
        case TokenType.REGEX:
            return "//"
        case TokenType.DIGIT:
            return r"\d"
        case TokenType.EXCLAMATION_MARK:
            return "!"
        case TokenType.DOT:
            return "."
        case TokenType.EQUALS:
            return "="
        case TokenType.SPACE:
            return " "
        case TokenType.PLAIN_TEXT:
            return raw.value
        # unrecognised token type
        case _:
            raise TwaddleParserException(
                f"[get_plain_text_for_object] unknown token type: {raw.type}"
            )


def to_plain_text_token_except(raw: Token, accept_list: tuple[TokenType]) -> Token:
    if raw.type in accept_list:
        return raw
    else:
        return to_plain_text_token(raw)


def merge_text_objects(raw: deque[TextObject]) -> TextObject:
    value = ""

    for token in raw:
        if isinstance(token, TextObject):
            value += token.text
        else:
            raise TwaddleParserException(
                f"[ParserUtils::merge_text_objects] object of type {type(token)} when RantTextObject was expected"
            )
    return TextObject(value)


def check_for_escape(next_token: Token) -> str | None:
    escapable_symbols = [
        TokenType.LEFT_ANGLE_BRACKET,
        TokenType.RIGHT_ANGLE_BRACKET,
        TokenType.LEFT_CURLY_BRACKET,
        TokenType.RIGHT_CURLY_BRACKET,
        TokenType.LEFT_SQUARE_BRACKET,
        TokenType.RIGHT_SQUARE_BRACKET,
        TokenType.PIPE,
    ]
    if next_token.type in escapable_symbols:
        return next_token
    return None



import pytest

from twaddle.exceptions import TwaddleFunctionException, TwaddleInterpreterException
from twaddle.interpreter.interpreter import Interpreter
from twaddle.lookup.lookup_manager import LookupManager


def get_standard_interpreter_output(sentence: str) -> str:
    interpreter = Interpreter(LookupManager())
    return interpreter.interpret_external(sentence)


def get_strict_interpreter_output(sentence: str) -> str:
    strict_interpreter = Interpreter(LookupManager(), strict_mode=True)
    return strict_interpreter.interpret_external(sentence)


def test_plain_text():
    hello_world = "hello, world"
    result = get_standard_interpreter_output(hello_world)
    assert result == hello_world


def test_text_with_special_characters():
    special_characters = "|-;::/:!.="
    result = get_standard_interpreter_output(special_characters)
    assert result == special_characters


def test_choice():
    result = get_standard_interpreter_output("{a|b}")
    assert result == "a" or result == "b"


def test_unknown_function_error():
    with pytest.raises(TwaddleInterpreterException) as e_info:
        get_standard_interpreter_output("[funk]")
    assert e_info.value.message == "[Interpreter::run] no function found named 'funk'"


def test_repeat():
    result = get_standard_interpreter_output("[rep:3]{a}")
    assert result == "aaa"


def test_nested_blocks():
    result = get_standard_interpreter_output("{{a|b}|{c|d}}")
    assert result in ["a", "b", "c", "d"]


# noinspection SpellCheckingInspection
def test_repeat_with_separator():
    result = get_standard_interpreter_output("[rep:3][sep:x]{a}")
    assert result == "axaxa"


def test_article_in_separator():
    result = get_standard_interpreter_output(r"[rep:2][sep: \a ]{egg}")
    assert result == "egg an egg"


# noinspection SpellCheckingInspection
def test_repeat_with_first_and_last():
    result = get_standard_interpreter_output("[rep:5][first:a][last:z]{x}")
    assert result == "axxxxzx"


def test_synchronizer_locked():
    result = get_standard_interpreter_output(
        "[x:tests;locked]{a|b|c}[x:tests]{a|b|c}[x:tests]{a|b|c}"
    )
    assert result in ["aaa", "bbb", "ccc"]


def test_synchronizer_deck():
    for _ in range(0, 10):
        result = get_standard_interpreter_output("[x:tests;deck]{a|b}[x:tests]{a|b}")
        assert result in ["ab", "ba"]


def test_random_number():
    for _ in range(0, 10):
        result_10 = int(get_standard_interpreter_output("[rand:0;10]"))
        result_negative = int(get_standard_interpreter_output("[rand:-10;-5]"))
        result_big = int(get_standard_interpreter_output("[rand:1000;2000]"))
        assert 0 <= result_10 <= 10
        assert -10 <= result_negative <= -5
        assert 1000 <= result_big <= 2000


def test_case():
    # result = get_interpreter_output("[case:upper]uPpEr [case:lower]loWeR")
    # assert result == "UPPER lower"
    # result = get_interpreter_output("[case:title]it's a title")
    # assert result == "It's A Title"
    result = get_standard_interpreter_output("\\a [case:title]egg")
    assert result == "an Egg"
    result = get_standard_interpreter_output(
        "[case:sentence]this is a sentence. this is another SENTENCE."
    )
    assert result == "This is a sentence. This is another sentence."


# noinspection SpellCheckingInspection
def test_case_block_interaction():
    result = get_standard_interpreter_output("look, a {dog[case:upper]}!")
    assert result == "look, a dog!"
    result = get_standard_interpreter_output("{a [case:upper]A|a [case:upper]A}")
    assert result == "a A"
    result = get_standard_interpreter_output("the {[case:title]A} team")
    assert result == "the A Team"
    result = get_standard_interpreter_output("[case:upper]The {cow|chicken}")
    assert result in ["THE COW", "THE CHICKEN"]
    result = get_standard_interpreter_output(
        "[case:title]The {[case:upper]BIG[case:lower]small}"
    )
    assert result == "The BIGsmall"


def test_digit():
    result = get_standard_interpreter_output(r"\d\d\d")
    for d in result:
        int(d)


# noinspection SpellCheckingInspection
def test_new_line():
    result = get_standard_interpreter_output(r"hello\nworld!")
    assert (
        result
        == """hello
world!"""
    )


def test_simple_regex():
    result = get_standard_interpreter_output("[//a//i:a bat;i]")
    assert result == "i bit"
    result = get_standard_interpreter_output("[//a//i:a;[match][match]]")
    assert result == "aa"


def test_hidden():
    result = get_standard_interpreter_output("[hide]{a secret}message")
    assert result == "message"


def test_indefinite_article_in_regex():
    result = get_standard_interpreter_output("[//[aeiou]//:i'm \\a cat;x]")
    assert result == "x'm x cxt"
    result = get_standard_interpreter_output("[//[aeiou]//:i'm \\a elephant;x]")
    assert result == "x'm xn xlxphxnt"


def test_gap_between_hide_function_and_block():
    result = get_standard_interpreter_output("[hide]something {and something else}")
    assert result == "something "


def test_gap_between_rep_function_and_block():
    result = get_standard_interpreter_output("[rep:2]something {repeating }")
    assert result == "something repeating repeating "


def test_separator_without_repetitions():
    result = get_standard_interpreter_output("[sep:x]{hey}")
    assert result == "hey"


def test_synchronizer_persistence():
    sync_interpreter = Interpreter(LookupManager(), persistent_synchronizers=True)

    sync_sentence = "[sync:test;locked]{a|b|c}"
    first_sync = sync_interpreter.interpret_external(sync_sentence)
    second_sync = sync_interpreter.interpret_external(sync_sentence)
    assert first_sync == second_sync


def test_clear_synchronizer_persistence():
    sync_interpreter = Interpreter(LookupManager(), persistent_synchronizers=True)
    results = [sync_interpreter.interpret_external("[sync:test;locked]{a|b|c}")]
    for _ in range(0, 10):
        sync_interpreter.force_clear()
        results.append(sync_interpreter.interpret_external("[sync:test;locked]{a|b|c}"))
    assert len(set(results)) > 1


def test_clear_function_in_sentence():
    sync_interpreter = Interpreter(LookupManager(), persistent_synchronizers=True)

    sentence = "[sync:test;locked]{a|b}[clear][sync:test;locked]{a|b}"
    results = [sync_interpreter.interpret_external(sentence)]
    for _ in range(0, 15):
        results.append(sync_interpreter.interpret_external(sentence))
    assert len(set(results)) > 2


def test_clear_synchronizers_with_persistence():
    sync_interpreter = Interpreter(LookupManager(), persistent_synchronizers=True)

    sentence = "[clear][sync:test;locked]{a|b} [sync:test]{a|b}"
    results = [sync_interpreter.interpret_external(sentence)]
    for _ in range(0, 15):
        results.append(sync_interpreter.interpret_external(sentence))
    assert len(set(results)) > 1


def test_synchronizers_normal_mode_not_when_num_choices_change():
    sync_types = ["locked", "cdeck", "deck"]
    for type in sync_types:
        initial_sync_block = f"[sync:test;{type}]"
        # don't care about response, just check no exception raised
        get_standard_interpreter_output(initial_sync_block + "{a|b} [sync:test]{a|b|c}")


def test_synchronizers_strict_mode_raise_when_num_choices_change():
    sync_types = ["locked", "cdeck", "deck"]
    for type in sync_types:
        with pytest.raises(TwaddleInterpreterException) as e_info:
            initial_sync_block = f"[sync:test;{type}]"
            get_strict_interpreter_output(
                initial_sync_block + "{a|b} [sync:test]{a|b|c}"
            )
        assert (
            e_info.value.message
            == "[Interpreter._get_synchronizer_for_block] Invalid number of choices (3) for"
            " synchronizer 'test', initialised with 2"
        )


def test_abbreviate_default_upper():
    assert (
        get_standard_interpreter_output(
            "[abbr]{Dennis obtains German geese in every situation}"
        )
        == "DOGGIES"
    )


def test_abbreviate_retain_case():
    assert (
        get_standard_interpreter_output(
            "[abbr:retain]{a boorish cat damaged everybody's feelings, going "
            "haywire in Japan, killing lemurs maniacally, never omitting personal quarrels, reactivating "
            "settled tensions until violently whacking xylophones yet zapped}"
        )
        == "abcdefghiJklmnopqrstuvwxyz"
    )


def test_abbreviate_lower_case():
    assert (
        get_standard_interpreter_output(
            "[abbr:lower]{Let ordinary Welshmen exist rationally}"
        )
        == "lower"
    )


def test_abbreviate_upper_case():
    assert (
        get_standard_interpreter_output(
            "[abbr:upper]{Unlikely, perhaps, plump Edwardian rabbits}"
        )
        == "UPPER"
    )


def test_abbreviate_first():
    assert (
        get_standard_interpreter_output(
            "[abbr:first]{friendly Ivan receives Spanish treats}"
        )
        == "First"
    )


def test_abbreviation_numbers():
    assert get_standard_interpreter_output("[abbr]{1234 Serious Items}") == "1234SI"


def test_abbreviation_invalid_case():
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_standard_interpreter_output("[abbr:dalj]{whatever}")
    assert (
        e_info.value.message
        == "[function_definitions#abbreviate] invalid case argument 'dalj'"
    )


if __name__ == "__main__":
    test_case()



from enum import Enum, auto


class TokenType(Enum):
    LEFT_ANGLE_BRACKET = auto()
    RIGHT_ANGLE_BRACKET = auto()
    LEFT_CURLY_BRACKET = auto()
    RIGHT_CURLY_BRACKET = auto()
    LEFT_SQUARE_BRACKET = auto()
    RIGHT_SQUARE_BRACKET = auto()
    PIPE = auto()
    HYPHEN = auto()
    SEMICOLON = auto()
    COLON = auto()
    DOUBLE_COLON = auto()
    QUOTE = auto()
    NEW_LINE = auto()
    TAB = auto()
    LOWER_INDEFINITE_ARTICLE = auto()
    UPPER_INDEFINITE_ARTICLE = auto()
    DOUBLE_BACKSLASH = auto()
    SINGLE_BACKSLASH = auto()
    FORWARD_SLASH = auto()
    REGEX = auto()
    DIGIT = auto()
    EXCLAMATION_MARK = auto()
    DOT = auto()
    EQUALS = auto()
    PLAIN_TEXT = auto()
    SPACE = auto()


class Token:
    def __init__(self, t: TokenType, val: str = ""):
        self.type = t
        self.value = val

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Token):
            return NotImplemented
        return self.type == o.type and self.value == o.value

    def __str__(self) -> str:
        return f"RantToken (type: {self.type}; value: {self.value})"



from typing import Optional

from twaddle.compiler.compiler_objects import RootObject
from twaddle.interpreter.formatting_object import FormattingStrategy


class BlockAttributes:
    def __init__(self):
        self.repetitions: int = 1
        self.separator: RootObject | None = None
        self.first: RootObject | None = None
        self.last: RootObject | None = None
        self.synchronizer: str | None = None
        self.synchronizer_type: str | None = None
        self.hidden: bool = False
        self.reverse: bool = False
        self.save_as: Optional[str] = None
        self.copy_as: Optional[str] = None
        self.abbreviate: bool = False
        self.abbreviation_case: Optional[FormattingStrategy] = None


class BlockAttributeManager:
    def __init__(self):
        self.saved_blocks = dict()
        self.current_attributes = BlockAttributes()

    def get_attributes(self) -> BlockAttributes:
        attributes = self.current_attributes
        self.current_attributes = BlockAttributes()
        return attributes

    def set_synchronizer(self, args: list[str]):
        self.current_attributes.synchronizer = args[0]
        if len(args) > 1:
            self.current_attributes.synchronizer_type = args[1]

    def save_block(self, name: str):
        self.current_attributes.save_as = name

    def copy_block(self, name: str):
        self.current_attributes.copy_as = name

    def clear(self):
        self.current_attributes = BlockAttributes()



from enum import Enum, auto
from typing import Type


class ObjectType(Enum):
    ROOT = auto()
    TEXT = auto()  # plain text
    LOOKUP = auto()  # something looked up from a dictionary
    BLOCK = auto()  # choice from multiple options
    FUNCTION = auto()  # call to a rant function
    INDEFINITE_ARTICLE = auto()  # a/an, depending what follows
    DIGIT = auto()  # a random digit 0-9
    REGEX = auto()  # a regex


class Object:
    def __init__(self, t: ObjectType):
        self.type = t


class RootObject(Object):
    def __init__(self):
        Object.__init__(self, ObjectType.ROOT)
        self.contents = list[Object]()

    def __getitem__(self, n: int):
        return self.contents[n]

    def __len__(self):
        return len(self.contents)

    def append(self, new_content: Object):
        self.contents.append(new_content)


class TextObject(Object):
    def __init__(self, text: str):
        Object.__init__(self, ObjectType.TEXT)
        self.text = text


class LookupObject(Object):
    def __init__(
        self,
        dictionary: str,
        form: str = None,
        positive_tags: set[str] = None,
        negative_tags: set[str] = None,
        positive_label: str = None,
        negative_labels: set[str] = None,
        strict_mode: bool = False,
    ):
        Object.__init__(self, ObjectType.LOOKUP)
        self.dictionary = dictionary
        self.form = form
        self.positive_tags = positive_tags
        self.negative_tags = negative_tags
        self.positive_label = positive_label
        self.negative_labels = negative_labels
        self.strict_mode = strict_mode


class BlockObject(Object):
    def __init__(self, choices: list[Type[RootObject]]):
        Object.__init__(self, ObjectType.BLOCK)
        self.choices = choices

    def __getitem__(self, n: int):
        return self.choices[n].contents

    def __len__(self):
        return len(self.choices)


class FunctionObject(Object):
    def __init__(self, func: str, args: list[RootObject]):
        Object.__init__(self, ObjectType.FUNCTION)
        self.func = func
        self.args = args


class RegexObject(Object):
    def __init__(self, regex: str, scope: RootObject, replacement: RootObject):
        Object.__init__(self, ObjectType.REGEX)
        self.regex = regex
        self.scope = scope
        self.replacement = replacement


class IndefiniteArticleObject(Object):
    def __init__(self, default_upper_case: bool = False):
        Object.__init__(self, ObjectType.INDEFINITE_ARTICLE)
        self.default_upper = default_upper_case


class DigitObject(Object):
    def __init__(self):
        Object.__init__(self, ObjectType.DIGIT)



import os

import pytest

from twaddle.exceptions import TwaddleLookupException
from twaddle.runner import TwaddleRunner


def relative_path_to_full_path(rel_path: str) -> str:
    current_dir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(current_dir, rel_path)


path = relative_path_to_full_path("../resources/valid_dicts")
standard_runner = TwaddleRunner(path)
strict_runner = TwaddleRunner(path, strict_mode=True)


def test_rant():
    assert standard_runner.run_sentence("hello world") == "hello world"


def test_lookups():
    assert standard_runner.run_sentence("hello <adj> world") == "hello happy world"
    assert (
        standard_runner.run_sentence(
            "hello, welcome to my <adj> <noun-building-!small>"
        )
        == "hello, welcome to my happy factory"
    )


def test_lookups_with_labels():
    assert (
        standard_runner.run_sentence(
            "big <noun-building-large::=a>, small <noun-building::!=a>"
        )
        == "big factory, small shed"
    )
    assert (
        standard_runner.run_sentence("my <noun-shape::=a> is a regular <noun::=a>")
        == "my hexagon is a regular hexagon"
    )


def test_persistent_lookups_with_labels():
    persistent_runner = TwaddleRunner(path, persistent=True)
    assert (
        persistent_runner.run_sentence(
            "big <noun-building-large::=a>, small <noun-building::!=a>"
        )
        == "big factory, small shed"
    )
    assert persistent_runner.run_sentence("<noun::=a>") == "factory"
    assert persistent_runner.run_sentence("<noun-building::!=a>") == "shed"


def test_clearing_lookup_labels():
    persistent_runner = TwaddleRunner(path, persistent=True)
    assert (
        persistent_runner.run_sentence(
            "big <noun-building-large::=a>, small <noun-building::!=a>"
        )
        == "big factory, small shed"
    )
    results = []
    for _ in range(0, 10):
        persistent_runner.clear()
        results.append(persistent_runner.run_sentence("<noun::=a>"))
    assert len(set(results)) > 1


def test_clearing_lookup_labels_in_sentence():
    persistent_runner = TwaddleRunner(path, persistent=True)
    assert (
        persistent_runner.run_sentence(
            "big <noun-building-large::=a>, small <noun-building::!=a>"
        )
        == "big factory, small shed"
    )
    results = []
    for _ in range(0, 10):
        results.append(persistent_runner.run_sentence("[clear]<noun::=a>"))
    assert len(set(results)) > 1
    results = []
    for _ in range(0, 10):
        results.append(
            persistent_runner.run_sentence(r"<noun-shape::=a>[clear]<noun::=a>")
        )
    assert len(set(results)) > 1


# noinspection SpellCheckingInspection
def test_repetition():
    assert (
        standard_runner.run_sentence("[rep:3]{<noun-shape>}") == "hexagonhexagonhexagon"
    )


def test_repetition_with_separator():
    assert standard_runner.run_sentence("[rep:3][sep: ]{hey}") == "hey hey hey"


def test_repetition_with_first():
    assert (
        standard_runner.run_sentence("[rep:3][sep:! ][first:give me ]{more}!")
        == "give me more! more! more!"
    )


def test_repetition_with_last():
    assert (
        standard_runner.run_sentence("[rep:3][sep:, ][last:, and ]{no}")
        == "no, no, and no"
    )


def test_repetition_with_article_in_args():
    assert (
        standard_runner.run_sentence(
            r"[rep:3][sep:, \a ][first:\a ][last:, and \a ]{<noun-vehicle>}"
        )
        == "an ambulance, an ambulance, and an ambulance"
    )


def test_indefinite_article():
    assert standard_runner.run_sentence("\\a bow and \\A arrow") == "a bow and An arrow"
    assert (
        standard_runner.run_sentence("[case:title]\\a bow and \\a arrow")
        == "A Bow And An Arrow"
    )
    assert (
        standard_runner.run_sentence("[case:upper]\\a bow and \\a arrow")
        == "A BOW AND AN ARROW"
    )


def test_sentence_case_with_first_person_pronoun():
    assert (
        standard_runner.run_sentence(
            "[case:sentence]i don't know if <pronoun> care, <pronoun.lower> guess i do"
        )
        == "I don't know if I care, i guess I do"
    )


def test_escaped_characters():
    assert standard_runner.run_sentence("\<hey\>\{\}") == "<hey>{}"


def test_label_only_persistence():
    label_only_runner = TwaddleRunner(path, persistent_labels=True)

    assert (
        label_only_runner.run_sentence(
            "big <noun-building-large::=a>, small <noun-building::!=a>"
        )
        == "big factory, small shed"
    )
    assert label_only_runner.run_sentence("<noun::=a>") == "factory"

    sync_sentence = "[sync:test;locked]{a|b|c}"
    sync_results = [label_only_runner.run_sentence(sync_sentence)]
    for _ in range(0, 10):
        sync_results.append(label_only_runner.run_sentence(sync_sentence))
    assert len(set(sync_results)) > 1


def test_synchronizer_only_persistence():
    sync_only_runner = TwaddleRunner(path, persistent_synchronizers=True)

    first = sync_only_runner.run_sentence("[sync:test;locked]{a|b|c}")
    second = sync_only_runner.run_sentence("[sync:test;locked]{a|b|c}")
    assert first == second

    sync_only_runner.run_sentence(
        "big <noun-building-large::=a>, small <noun-building::!=a>"
    )
    results = [sync_only_runner.run_sentence("<noun::=a>")]
    for _ in range(0, 10):
        results.append(sync_only_runner.run_sentence("<noun::=a>"))
    assert len(set(results)) > 1


def test_persistence_overrides():
    runner = TwaddleRunner(
        path, persistent=True, persistent_labels=False, persistent_synchronizers=False
    )

    assert (
        runner.run_sentence("big <noun-building-large::=a>, small <noun-building::!=a>")
        == "big factory, small shed"
    )
    assert runner.run_sentence("<noun::=a>") == "factory"

    first = runner.run_sentence("[sync:test;locked]{a|b|c}")
    second = runner.run_sentence("[sync:test;locked]{a|b|c}")
    assert first == second


def test_patter_persistence():
    pattern_persistent_runner = TwaddleRunner(path, persistent_patterns=True)

    first = pattern_persistent_runner.run_sentence("[save:a]{<noun-vehicle>}")
    second = pattern_persistent_runner.run_sentence("[load:a]")

    assert first == second


# noinspection SpellCheckingInspection,PyPep8
def test_regex():
    assert standard_runner.run_sentence("[//a//i:a;\\a <noun-shape>]") == "a hexagon"
    assert (
        standard_runner.run_sentence("[//hexagon//:hexagon;a [match] has 6 sides]")
        == "a hexagon has 6 sides"
    )
    assert (
        standard_runner.run_sentence("[//^\w\w[aou]?//i:this;{[match]tab}]")
        == "thtabis"
    )
    assert (
        standard_runner.run_sentence(
            "[//tab.*//i:[//^\w\w[aou]?//i:this;{[match]tab}];tab]"
        )
        == "thtab"
    )


def test_assign_tags_in_hidden_block():
    assert (
        standard_runner.run_sentence(
            r"[hide]{<noun-shape::=a><noun-animal::=b><noun-vehicle::=c>}"
            "<noun::=b> <noun::=a> <noun::=c>"
        )
        == "dog hexagon ambulance"
    )


# noinspection SpellCheckingInspection,PyPep8
def test_complex_sentence():
    assert (
        standard_runner.run_sentence("[rep:2]{[//[aeiou]//:<noun-shape>;o]}")
        == "hoxogonhoxogon"
    )
    assert (
        standard_runner.run_sentence(
            "[case:title]the <noun-building-small::=a> and \\a [//hat//:hat;<noun-building::!=a>]"
        )
        == "The Shed And A Factory"
    )
    assert (
        standard_runner.run_sentence(
            "[//s\w//:suspicious slimy slithery snakes;ss[match]]"
        )
        == "sssussspicious ssslimy ssslithery sssnakes"
    )
    assert (
        standard_runner.run_sentence("[//[3-5]//:[rep:3][sep:\\n]{123456};x]")
        == """12xxx6
12xxx6
12xxx6"""
    )


def test_indefinite_article_at_block_end():
    assert standard_runner.run_sentence("{\\a} cat and {\\a} egg") == "a cat and an egg"


def test_indefinite_article_from_lookup():
    assert (
        standard_runner.run_sentence(
            "[case:sentence]<article.indefinite> cat and <article.indefinite> aardvark"
        )
        == "A cat and an aardvark"
    )


def test_strict_mode_error_invalid_class():
    with pytest.raises(TwaddleLookupException) as e_info:
        strict_runner.run_sentence("<noun-notarealclass>")
    assert (
        e_info.value.message
        == "[LookupDictionary._strict_class_validation] Invalid class 'notarealclass' requested "
        "for dictionary 'noun' in strict mode"
    )


def test_strict_mode_undefined_label():
    with pytest.raises(TwaddleLookupException) as e_info:
        strict_runner.run_sentence("<noun::!=undefined>")
    assert (
        e_info.value.message
        == "[LookupDictionary._strict_label_validation] Requested antimatch of label "
        "'undefined', not defined for dictionary 'noun'"
    )


def test_reverse_with_lookups():
    assert strict_runner.run_sentence("[reverse]{<noun-vehicle>}") == "ecnalubma"


def test_save_load_block_with_lookups():
    assert (
        standard_runner.run_sentence("[save:a]{<noun-vehicle>} [load:a]")
        == "ambulance ambulance"
    )


def test_save_load_block_with_reverse():
    assert (
        standard_runner.run_sentence("[save:a]{<noun-vehicle>} [reverse]{[load:a]}")
        == "ambulance ecnalubma"
    )


def test_save_load_with_case():
    assert (
        standard_runner.run_sentence(
            "[save:a]{<noun-vehicle>} [case:upper][reverse]{[load:a]}"
        )
        == "ambulance ECNALUBMA"
    )


def test_save_load_with_synchronizer():
    assert standard_runner.run_sentence("[save:a]{[sync:x;deck]{a|b}} [load:a]") in [
        "a b",
        "b a",
    ]


def test_save_effective_before_reverse():
    for sentence in [
        "[save:a][reverse]{abc} [load:a]",
        "[reverse][save:a]{abc} [load:a]",
        "[reverse]{[save:a]{abc}} [load:a]",
        "[save:a]{[reverse]{abc}} [reverse]{[load:a]}",
    ]:
        assert standard_runner.run_sentence(sentence) == "cba abc"


def test_copy_paste_with_lookups():
    assert (
        standard_runner.run_sentence("[copy:a]{<noun-vehicle>} [paste:a]")
        == "ambulance ambulance"
    )


def test_copy_load_block_with_reverse():
    assert (
        standard_runner.run_sentence("[copy:a]{<noun-vehicle>} [reverse]{[paste:a]}")
        == "ambulance ecnalubma"
    )


def test_copy_load_with_case():
    assert (
        standard_runner.run_sentence(
            "[copy:a]{<noun-vehicle>} [case:upper][reverse]{[paste:a]}"
        )
        == "ambulance ECNALUBMA"
    )


def test_copy_paste_with_synchronizer():
    assert standard_runner.run_sentence("[copy:a]{[sync:x;deck]{a|b}} [paste:a]") in [
        "a a",
        "b b",
    ]


def test_copy_effective_before_reverse():
    for sentence in [
        "[copy:a][reverse]{abc} [paste:a]",
        "[reverse][copy:a]{abc} [paste:a]",
        "[reverse]{[copy:a]{abc}} [paste:a]",
        "[copy:a]{[reverse]{abc}} [reverse]{[paste:a]}",
    ]:
        assert standard_runner.run_sentence(sentence) == "cba abc"



from twaddle.interpreter.formatter import Formatter
from twaddle.interpreter.formatting_object import FormattingStrategy

formatter = Formatter()


def test_simple_print():
    formatter.append("hello")
    assert formatter.resolve() == "hello"


def test_normal_upper():
    formatter.append("abc")
    formatter.set_strategy(FormattingStrategy.UPPER)
    formatter.append("def")
    assert formatter.resolve() == "abcDEF"


# noinspection SpellCheckingInspection
def test_upper_lower():
    formatter.append("abc")
    formatter.set_strategy(FormattingStrategy.UPPER)
    formatter.append("def")
    formatter.set_strategy(FormattingStrategy.LOWER)
    formatter.append("GHI")
    assert formatter.resolve() == "abcDEFghi"


def test_sentence():
    formatter.set_strategy(FormattingStrategy.SENTENCE)
    formatter.append("hey there! this is. a test")
    assert formatter.resolve() == "Hey there! This is. A test"


def test_sentence_with_i():
    formatter.set_strategy(FormattingStrategy.SENTENCE)
    formatter.append("will this work how I expect?")
    assert formatter.resolve() == "Will this work how I expect?"


def test_i_edge_cases():
    formatter.set_strategy(FormattingStrategy.SENTENCE)
    formatter.append("ii i5 i! i i I i.")
    assert formatter.resolve() == "Ii i5 I! I I I I."


def test_title():
    formatter.set_strategy(FormattingStrategy.TITLE)
    formatter.append("hey there! this text's a test")
    assert formatter.resolve() == "Hey There! This Text's A Test"


if __name__ == "__main__":
    test_sentence()



import pytest

from twaddle.compiler.compiler import Compiler, CompilerContext, CompilerContextStack
from twaddle.compiler.compiler_objects import (
    BlockObject,
    FunctionObject,
    IndefiniteArticleObject,
    LookupObject,
    ObjectType,
    RegexObject,
    RootObject,
    TextObject,
)
from twaddle.exceptions import TwaddleParserException

standard_compiler = Compiler()
strict_compiler = Compiler(strict_mode=True)


def get_standard_compile_result(sentence: str) -> RootObject:
    return standard_compiler.compile(sentence).contents


def get_strict_compile_result(sentence: str) -> RootObject:
    return strict_compiler.compile(sentence).contents


def test_compiler_context_stack():
    stack = CompilerContextStack()
    assert stack.current_context() == CompilerContext.ROOT
    stack.add_context(CompilerContext.FUNCTION)
    stack.add_context(CompilerContext.BLOCK)
    assert stack.current_context() == CompilerContext.BLOCK
    stack.remove_context(CompilerContext.BLOCK)
    assert stack.current_context() == CompilerContext.FUNCTION
    with pytest.raises(TwaddleParserException) as e_info:
        stack.remove_context(CompilerContext.BLOCK)
    assert (
        e_info.value.message
        == "[CompilerContextStack::remove_context] tried to remove BLOCK but current context is FUNCTION"
    )


def test_parse_text():
    result = get_standard_compile_result("hello")
    assert len(result) == 1
    assert result[0].type == ObjectType.TEXT
    assert result[0].text == "hello"


def test_parse_simple_lookup():
    result = get_standard_compile_result("<whatever>")
    assert len(result) == 1
    lookup: LookupObject = result[0]
    assert lookup.type == ObjectType.LOOKUP
    assert lookup.dictionary == "whatever"


def test_parse_complex_lookups():
    result = get_standard_compile_result("<dictionary.form-category>")
    assert len(result) == 1
    lookup: LookupObject = result[0]
    assert lookup.type == ObjectType.LOOKUP
    assert lookup.dictionary == "dictionary"
    assert lookup.form == "form"
    assert lookup.positive_tags == {"category"}
    result = get_standard_compile_result("<dictionary.form-!category::=a>")
    lookup = result[0]
    assert len(lookup.positive_tags) == 0
    assert lookup.negative_tags == {"category"}
    assert lookup.positive_label == "a"
    result = get_standard_compile_result(
        "<dictionary-category1-category2-!category3::!=b>"
    )
    lookup = result[0]
    assert lookup.positive_tags == {"category1", "category2"}
    assert lookup.negative_tags == {"category3"}
    assert lookup.positive_label is None
    assert lookup.negative_labels == {"b"}


def test_parse_function():
    lex_result = get_standard_compile_result("[function:arg1;arg2]")
    assert len(lex_result) == 1
    assert isinstance(lex_result[0], FunctionObject)
    func: FunctionObject = lex_result[0]
    assert func.func == "function"
    assert len(func.args) == 2
    assert func.args[0][0].text == "arg1"
    assert func.args[1][0].text == "arg2"


def test_parse_choice():
    parser_output = get_standard_compile_result("{this|that}")
    assert len(parser_output) == 1
    assert isinstance(parser_output[0], BlockObject)
    choice_result: BlockObject = parser_output[0]
    assert choice_result.type == ObjectType.BLOCK
    assert len(choice_result.choices) == 2
    for choice in choice_result.choices:
        assert len(choice.contents) == 1
    assert choice_result.choices[0][0].text == "this"
    assert choice_result.choices[1][0].text == "that"


def test_nested_block():
    parser_output = get_standard_compile_result("{{a|b}|{c|d}}")
    assert len(parser_output) == 1
    outer_block = parser_output[0]
    assert isinstance(outer_block, BlockObject)
    ab: BlockObject = outer_block[0][0]
    cd: BlockObject = outer_block[1][0]
    assert len(ab) == 2
    assert len(cd) == 2
    assert ab[0][0].text == "a"
    assert ab[1][0].text == "b"
    assert cd[0][0].text == "c"
    assert cd[1][0].text == "d"


def test_parse_choice_with_lookups():
    parser_output = get_standard_compile_result("{this|<something.form-category>}")
    assert len(parser_output) == 1
    assert isinstance(parser_output[0], BlockObject)
    block: BlockObject = parser_output[0]
    assert isinstance(block[0][0], TextObject)
    assert block[0][0].text == "this"
    assert isinstance(block[1][0], LookupObject)
    lookup: LookupObject = block[1][0]
    assert lookup.dictionary == "something"
    assert lookup.form == "form"
    assert lookup.positive_tags == {"category"}


def test_parse_indefinite_article():
    parser_output = get_standard_compile_result("\\a bow and \\a arrow")
    assert len(parser_output) == 4
    assert isinstance(parser_output[0], IndefiniteArticleObject)
    assert isinstance(parser_output[1], TextObject)
    assert parser_output[1].text == " bow and "
    assert isinstance(parser_output[2], IndefiniteArticleObject)
    assert isinstance(parser_output[3], TextObject)
    assert parser_output[3].text == " arrow"


def test_parse_escaped_characters():
    parser_output = get_standard_compile_result(r"\<angles\>")
    assert len(parser_output) == 3
    for index in [0, 1, 2]:
        assert isinstance(parser_output[index], TextObject)
    assert parser_output[0].text == "<"
    assert parser_output[1].text == "angles"
    assert parser_output[2].text == ">"


def test_parse_article_in_args():
    parser_output = get_standard_compile_result(r"[rep:2][sep:\a]{<noun>}")
    separator_args = parser_output[1].args
    assert isinstance(separator_args[0].contents[0], IndefiniteArticleObject)


def test_parse_simple_regex():
    parser_output = get_standard_compile_result("[//a//i:a bat;i]")
    assert len(parser_output) == 1
    rro: RegexObject = parser_output[0]
    assert rro.regex == "a"
    assert rro.scope[0].text == "a bat"
    assert rro.replacement[0].text == "i"


# noinspection SpellCheckingInspection,PyPep8
def test_parse_complex_regex():
    parser_output = get_standard_compile_result(
        "[//^\w\w[aeiou]?//i:whatever;something]"
    )
    assert len(parser_output) == 1
    rro: RegexObject = parser_output[0]
    assert rro.regex == "^\w\w[aeiou]?"


def test_strict_mode():
    result = get_strict_compile_result("<dict-tag.form>")
    assert len(result) == 1
    lookup: LookupObject = result[0]
    assert lookup.strict_mode


def test_raise_on_invalid_closing_brackets():
    for bracket in ["]", ">", "}"]:
        with pytest.raises(TwaddleParserException) as e_info:
            get_standard_compile_result(bracket)
        assert e_info.value is not None
        with pytest.raises(TwaddleParserException) as e_info:
            get_standard_compile_result("some text here" + bracket)
        assert e_info.value is not None


if __name__ == "__main__":
    test_parse_complex_regex()



from .function_definitions import (
    abbreviate,
    case,
    copy,
    first,
    hide,
    last,
    match,
    rand,
    repeat,
    reverse,
    save,
    separator,
    sync,
)

function_definitions = {
    "abbr": abbreviate,
    "abbreviate": abbreviate,
    "case": case,
    "copy": copy,
    "first": first,
    "hide": hide,
    "last": last,
    "match": match,
    "rand": rand,
    "rep": repeat,
    "reverse": reverse,
    "save": save,
    "sep": separator,
    "x": sync,
    "sync": sync,
}



from collections import deque
from enum import Enum, auto

from twaddle.exceptions import TwaddleParserException
from twaddle.lexer.lexer import lex
from twaddle.lexer.lexer_tokens import Token, TokenType

from .compiler_objects import (
    BlockObject,
    DigitObject,
    FunctionObject,
    IndefiniteArticleObject,
    LookupObject,
    Object,
    RegexObject,
    RootObject,
    TextObject,
)
from .compiler_utils import check_for_escape, get_text_for_object, to_plain_text_object


class CompilerContext(Enum):
    ROOT = auto()
    FUNCTION = auto()
    LOOKUP = auto()
    BLOCK = auto()
    REGEX = auto()


class CompilerContextStack:
    def __init__(self):
        self.stack = deque[CompilerContext]()
        self.stack.append(CompilerContext.ROOT)

    def current_context(self):
        return self.stack[-1]

    def add_context(self, context: CompilerContext):
        self.stack.append(context)

    def remove_context(self, context: CompilerContext):
        if self.current_context() is not context:
            raise TwaddleParserException(
                f"[CompilerContextStack::remove_context] tried to remove {context.name} "
                f"but current context is {self.stack[-1].name}"
            )
        self.stack.pop()


class Compiler:
    def __init__(self, strict_mode: bool = False):
        self.context = CompilerContextStack()
        self.strict_mode = strict_mode

    def compile(self, sentence: str) -> RootObject:
        self.context = CompilerContextStack()
        result = self.parse_root(lex(sentence))
        if self.context.current_context() is not CompilerContext.ROOT:
            unexpected_context = self.context.current_context()
            raise TwaddleParserException(
                f"[RantCompiler::compile] reached end while still in {unexpected_context.name} context"
            )
        return result

    def parse_root(self, tokens: deque[Token]) -> RootObject:
        result = RootObject()
        while tokens:
            context = self.context.current_context()
            token = tokens[0]
            match token.type:
                case TokenType.PLAIN_TEXT:
                    result.append(TextObject(token.value))
                    tokens.popleft()
                case TokenType.LEFT_ANGLE_BRACKET:
                    self.context.add_context(CompilerContext.LOOKUP)
                    result.append(self.parse_lookup(tokens))
                case TokenType.LEFT_CURLY_BRACKET:
                    self.context.add_context(CompilerContext.BLOCK)
                    result.append(self.parse_block(tokens))
                case TokenType.LEFT_SQUARE_BRACKET:
                    if len(tokens) and tokens[1].type == TokenType.REGEX:
                        self.context.add_context(CompilerContext.REGEX)
                        result.append(self.parse_regex(tokens))
                    else:
                        self.context.add_context(CompilerContext.FUNCTION)
                        result.append(self.parse_function(tokens))
                case TokenType.PIPE:
                    if context is CompilerContext.BLOCK:
                        return result
                    else:
                        result.append(to_plain_text_object(token))
                        tokens.popleft()
                case TokenType.RIGHT_CURLY_BRACKET:
                    if context is CompilerContext.BLOCK:
                        return result
                    raise TwaddleParserException(
                        "[Compiler.parse_root] Unexpected right angle bracket"
                    )
                case TokenType.COLON:
                    if context is CompilerContext.FUNCTION:
                        return result
                    if context is CompilerContext.REGEX:
                        return result
                    else:
                        result.append(to_plain_text_object(token))
                        tokens.popleft()
                case TokenType.SEMICOLON:
                    if context is CompilerContext.FUNCTION:
                        return result
                    if context is CompilerContext.REGEX:
                        return result
                    else:
                        result.append(to_plain_text_object(token))
                        tokens.popleft()
                case TokenType.RIGHT_SQUARE_BRACKET:
                    if context is CompilerContext.FUNCTION:
                        return result
                    if context is CompilerContext.REGEX:
                        return result
                    raise TwaddleParserException(
                        "[Compiler.parse_root] Unexpected right square bracket"
                    )
                case TokenType.RIGHT_ANGLE_BRACKET:
                    if context is CompilerContext.LOOKUP:
                        return result
                    raise TwaddleParserException(
                        "[Compiler.parse_root] Unexpected right angle bracket"
                    )
                case TokenType.LOWER_INDEFINITE_ARTICLE:
                    result.append(IndefiniteArticleObject())
                    tokens.popleft()
                case TokenType.UPPER_INDEFINITE_ARTICLE:
                    result.append(IndefiniteArticleObject(default_upper_case=True))
                    tokens.popleft()
                case TokenType.DIGIT:
                    result.append(DigitObject())
                    tokens.popleft()
                case TokenType.DOUBLE_BACKSLASH:
                    result.append(to_plain_text_object(token))
                    tokens.popleft()
                case TokenType.SINGLE_BACKSLASH:
                    tokens.popleft()
                    if len(tokens) and (escaped := check_for_escape(tokens[0])):
                        result.append(to_plain_text_object(escaped))
                        tokens.popleft()
                    else:
                        result.append(to_plain_text_object(token))
                # more special cases to handle here later, just convert to text for now
                case _:
                    result.append(to_plain_text_object(token))
                    tokens.popleft()
        return result

    def parse_lookup(self, tokens: deque[Token]) -> Object:
        form = None
        positive_tags = set[str]()
        negative_tags = set[str]()
        positive_label = None
        negative_labels = set[str]()

        # first thing must always be the opening angle bracket:
        if tokens[0].type is not TokenType.LEFT_ANGLE_BRACKET:
            raise TwaddleParserException(
                "[Compiler.parse_block] input does not begin with left angle bracket"
            )
        tokens.popleft()

        # next thing must always be the dictionary name, so it has to be text:
        if tokens[0].type is not TokenType.PLAIN_TEXT:
            raise TwaddleParserException(
                "[Compiler.parse_block] opening angle bracket must be followed by dictionary name"
            )

        # read the dictionary name and get rid of it so we can deal with the less fixed stuff
        dictionary = tokens.popleft().value

        while len(tokens) > 0:
            token = tokens.popleft()
            match token.type:
                case TokenType.RIGHT_ANGLE_BRACKET:
                    self.context.remove_context(CompilerContext.LOOKUP)
                    return LookupObject(
                        dictionary,
                        form,
                        positive_tags,
                        negative_tags,
                        positive_label,
                        negative_labels,
                        strict_mode=self.strict_mode,
                    )
                case TokenType.DOT:
                    if tokens[0].type is not TokenType.PLAIN_TEXT:
                        raise TwaddleParserException(
                            "[Compiler.parse_block] dot must be followed by form"
                        )
                    form = tokens.popleft().value
                    continue
                case TokenType.HYPHEN:
                    positive = True
                    if tokens[0].type is TokenType.EXCLAMATION_MARK:
                        tokens.popleft()
                        positive = False
                    if tokens[0].type is not TokenType.PLAIN_TEXT:
                        raise TwaddleParserException(
                            "[Compiler.parse_block] hyphen must be followed by category"
                        )
                    category = tokens.popleft().value
                    if positive:
                        positive_tags.add(category)
                    else:
                        negative_tags.add(category)
                    continue
                case TokenType.DOUBLE_COLON:
                    if tokens[0].type is TokenType.EQUALS:
                        tokens.popleft()
                        if len(tokens) > 0 and tokens[0].type is TokenType.PLAIN_TEXT:
                            positive_label = tokens.popleft().value
                        else:
                            raise TwaddleParserException(
                                "[Compiler.parse_block] no valid definition for match"
                            )
                    elif tokens[0].type is TokenType.EXCLAMATION_MARK:
                        tokens.popleft()
                        if (
                            len(tokens) >= 2
                            and tokens[0].type is TokenType.EQUALS
                            and tokens[1].type is TokenType.PLAIN_TEXT
                        ):
                            # get rid of the equals
                            tokens.popleft()
                            # get label name
                            negative_labels.add(tokens.popleft().value)
                        else:
                            raise TwaddleParserException(
                                "[Compiler.parse_block] no valid definition for anti-match"
                            )
                case _:
                    continue
        # if we reach here, something went wrong
        raise TwaddleParserException(
            "[Compiler.parse_block] Error parsing dictionary lookup, probably an invalid character"
        )

    def parse_block(self, tokens: deque[Token]) -> BlockObject:
        choices = list()
        this_choice = RootObject()
        if tokens[0].type is not TokenType.LEFT_CURLY_BRACKET:
            raise TwaddleParserException(
                "[Compiler.parse_block] block factory called without '{', this shouldn't happen!"
            )
        tokens.popleft()
        while len(tokens) > 0:
            token = tokens[0]
            match token.type:
                case TokenType.PIPE:
                    tokens.popleft()
                    choices.append(this_choice)
                    continue
                case TokenType.RIGHT_CURLY_BRACKET:
                    tokens.popleft()
                    choices.append(this_choice)
                    self.context.remove_context(CompilerContext.BLOCK)
                    return BlockObject(choices)
                case _:
                    this_choice = self.parse_root(tokens)
        # something went wrong, fall over
        raise TwaddleParserException(
            "[Compiler.parse_block] something went wrong, probably a missing '}'"
        )

    # noinspection GrazieInspection
    def parse_function(self, tokens: deque[Token]) -> FunctionObject:
        args = list()

        # first thing must always be the opening square bracket:
        if tokens[0].type is not TokenType.LEFT_SQUARE_BRACKET:
            raise TwaddleParserException(
                "[Compiler.parse_function] input does not begin with left angle bracket"
            )
        tokens.popleft()

        # next thing must always be the function name, so it has to be text:
        if tokens[0].type is not TokenType.PLAIN_TEXT:
            raise TwaddleParserException(
                "[Compiler.parse_function] expected function name"
            )

        # read the dictionary name and get rid of it so we can deal with the less fixed stuff
        func = tokens.popleft().value

        while len(tokens) > 0:
            token = tokens.popleft()
            match token.type:
                case TokenType.COLON:
                    args.append(self.parse_root(tokens))
                    continue
                case TokenType.SEMICOLON:
                    args.append(self.parse_root(tokens))
                    continue
                case TokenType.RIGHT_SQUARE_BRACKET:
                    self.context.remove_context(CompilerContext.FUNCTION)
                    return FunctionObject(func, args)
                case _:
                    continue
        # if we reach here, something went wrong
        raise TwaddleParserException(
            "[Compiler.parse_function] Error parsing function, probably an invalid character or missing closing bracket"
        )

    def parse_regex(self, tokens: deque[Token]) -> RegexObject:
        regex = ""
        scope = RootObject()
        replacement = RootObject()

        # first thing must always be the opening square bracket:
        if tokens[0].type is not TokenType.LEFT_SQUARE_BRACKET:
            raise TwaddleParserException(
                "[Compiler.parse_regex] input does not begin with left angle bracket"
            )
        tokens.popleft()

        # next thing must always be the regex
        if tokens[0].type is not TokenType.REGEX:
            raise TwaddleParserException(
                "[Compiler.parse_regex] expected function name"
            )
        tokens.popleft()

        while len(tokens) and tokens[0].type is not TokenType.REGEX:
            token = tokens.popleft()
            match token.type:
                case TokenType.PLAIN_TEXT:
                    regex += token.value
                case _:
                    regex += get_text_for_object(token)

        if len(tokens) == 0:
            raise TwaddleParserException(
                "[Compiler.parse_regex] reached end of input without finding end of regex"
            )
        tokens.popleft()

        while len(tokens):
            token = tokens.popleft()
            match token.type:
                case TokenType.COLON:
                    scope = self.parse_root(tokens)
                case TokenType.SEMICOLON:
                    replacement = self.parse_root(tokens)
                case TokenType.RIGHT_SQUARE_BRACKET:
                    self.context.remove_context(CompilerContext.REGEX)
                    return RegexObject(regex, scope, replacement)
                case _:
                    continue
        # if we reach here, something went wrong
        raise TwaddleParserException(
            "[Compiler.parse_regex] Error parsing regex, probably an invalid character or missing closing bracket"
        )



from twaddle.compiler.compiler_objects import (
    BlockObject,
    FunctionObject,
    LookupObject,
    ObjectType,
    TextObject,
)


def test_rant_text_object():
    rt = TextObject("hello")
    assert rt.type == ObjectType.TEXT
    assert rt.text == "hello"


def test_rant_lookup_object():
    rl = LookupObject("dictionary", "form")
    assert rl.type == ObjectType.LOOKUP
    assert rl.dictionary == "dictionary"
    assert rl.form == "form"


def test_rant_choice_object():
    r1 = TextObject("hello")
    r2 = LookupObject("name")
    choices = [r1, r2]
    rc = BlockObject(choices)
    assert rc.type == ObjectType.BLOCK
    assert len(rc.choices) == 2
    assert rc.choices[0] == r1
    assert rc.choices[1] == r2


def test_rant_function_object():
    rf = FunctionObject("function", [TextObject("arg0"), TextObject("arg1")])
    assert rf.type == ObjectType.FUNCTION
    assert rf.func == "function"
    assert len(rf.args) == 2
    assert rf.args[0].text == "arg0"
    assert rf.args[1].text == "arg1"



from twaddle.interpreter.interpreter import Interpreter
from twaddle.lookup.lookup_manager import LookupManager


def get_interpreter_output(sentence: str) -> str:
    interpreter = Interpreter(LookupManager())
    return interpreter.interpret_external(sentence)


def test_reverse_basic():
    result = get_interpreter_output("[reverse]{hello world}")
    assert result == "dlrow olleh"


def test_reverse_with_nested_functions():
    result = get_interpreter_output("[reverse][case:upper]{hello}")
    assert result == "OLLEH"


def test_reverse_empty_block():
    result = get_interpreter_output("[reverse]{}")
    assert result == ""


def test_reverse_with_punctuation():
    result = get_interpreter_output("[reverse]{hello, world!}")
    assert result == "!dlrow ,olleh"


def test_reverse_with_nested_blocks():
    result = get_interpreter_output("[reverse]{[rep:2][sep:\s]{hello}}")
    assert result == "olleh olleh"


def test_reverse_special_characters():
    result = get_interpreter_output("[reverse]{hello\nworld}")
    assert result == "dlrow\nolleh"


def test_reverse_unicode():
    result = get_interpreter_output("[reverse]{héllö wörld}")
    assert result == "dlröw ölléh"



import os
from pathlib import Path

import pytest

from twaddle.compiler.compiler_objects import IndefiniteArticleObject
from twaddle.exceptions import TwaddleDictionaryException, TwaddleLookupException
from twaddle.lookup.dictionary_file_parser import DictionaryFileParser
from twaddle.lookup.lookup_dictionary import LookupDictionary, LookupObject
from twaddle.lookup.lookup_entry import DictionaryEntry
from twaddle.lookup.lookup_manager import LookupManager


def relative_path_to_full_path(rel_path: str) -> str:
    current_dir = os.path.dirname(os.path.realpath(__file__))
    return Path(os.path.join(current_dir, rel_path))


# noinspection SpellCheckingInspection
def test_lookup_type():
    lookup_thing = DictionaryEntry(
        {
            "singular": "thing",
            "plural": "things",
            "possessive": "thing's",
            "pluralpossessive": "things'",
        }
    )
    assert lookup_thing["singular"] == "thing"
    assert lookup_thing["plural"] == "things"
    assert lookup_thing["possessive"] == "thing's"
    assert lookup_thing["pluralpossessive"] == "things'"


def test_dictionary():
    dictionary = LookupDictionary("noun", ["singular", "plural"])
    dictionary.add(["hexagon", "hexagons"])
    assert len(dictionary.entries) == 1
    lookup = LookupObject(dictionary="noun", form="singular")
    assert dictionary._get(lookup) == "hexagon"
    lookup.form = "plural"
    assert dictionary._get(lookup) == "hexagons"
    with pytest.raises(TwaddleLookupException) as e_info:
        lookup.form = "invalid"
        dictionary._get(lookup)
    assert (
        e_info.value.message
        == "[LookupDictionary.get] dictionary 'noun' has no form 'invalid'"
    )


def test_tag_requirement():
    dictionary = LookupDictionary("noun", ["singular", "plural"])
    dictionary.add(["thing", "things"], {"tag1"})
    dictionary.add(["hexagon", "hexagons"], {"tag2"})
    for _ in range(0, 5):
        lookup = LookupObject("noun")
        lookup.form = "plural"
        lookup.positive_tags = {"tag1"}
        assert dictionary._get(lookup) == "things"
        lookup.form = "singular"
        lookup.positive_tags = {"tag2"}
        assert dictionary._get(lookup) == "hexagon"
        lookup.form = "plural"
        lookup.positive_tags = {}
        lookup.negative_tags = {"tag1"}
        assert dictionary._get(lookup) == "hexagons"


def test_label_positive():
    dictionary = LookupDictionary("noun", ["singular", "plural"])
    dictionary.add(["thing", "things"], {"tag1"})
    dictionary.add(["hexagon", "hexagons"], {"tag2"})
    lookup = LookupObject(
        dictionary="noun",
        form="singular",
        positive_tags={"tag1"},
        positive_label="tests",
    )
    assert dictionary._get(lookup) == "thing"
    for _ in range(0, 5):
        lookup.positive_tags = {}
        assert dictionary._get(lookup) == "thing"


def test_labels_negative():
    dictionary = LookupDictionary("noun", ["singular", "plural"])
    dictionary.add(["thing", "things"], {"tag1"})
    dictionary.add(["hexagon", "hexagons"], {"tag2"})
    positive_tag_lookup = LookupObject(
        dictionary="noun",
        form="singular",
        positive_tags={"tag1"},
        positive_label="tests",
    )
    assert dictionary._get(positive_tag_lookup) == "thing"
    for _ in range(0, 5):
        negative_tag_lookup = LookupObject(
            dictionary="noun", form="singular", negative_labels={"tests"}
        )
        assert dictionary._get(negative_tag_lookup) == "hexagon"
        # just to check no problems with undefined labels
        negative_tag_lookup.negative_labels = {"hat"}
        assert dictionary._get(negative_tag_lookup)
    dictionary.clear_labels()
    results_after_clearing = list[str]()
    for _ in range(0, 50):
        negative_tag_lookup = LookupObject(
            dictionary="noun", form="singular", negative_labels={"tests"}
        )
        results_after_clearing.append(dictionary._get(negative_tag_lookup))
    assert "thing" in results_after_clearing


def test_dictionary_attributes_from_lines():
    factory = DictionaryFileParser()
    name = factory.read_name("#name noun")
    assert name == "noun"
    forms = factory.read_forms("#subs singular plural")
    assert forms == ["singular", "plural"]
    forms = factory.read_forms("#forms singular plural")
    assert forms == ["singular", "plural"]


def test_dictionary_read_from_file_simple():
    factory = DictionaryFileParser()
    path = relative_path_to_full_path("../resources/valid_dicts/example.dic")
    dictionary = factory.read_from_path(path)
    lookup = LookupObject("adj")
    assert dictionary.name == "adj"
    assert dictionary.forms == ["adj", "ness"]
    assert dictionary._get(lookup) == "happy"
    lookup.form = "ness"
    assert dictionary._get(lookup) == "happiness"
    lookup.form = None
    assert dictionary._get(lookup) == "happy"


def test_dictionary_read_from_file_with_classes():
    factory = DictionaryFileParser()
    path = relative_path_to_full_path(
        "../resources/valid_dicts/example_with_classes.dic"
    )
    dictionary = factory.read_from_path(path)
    assert dictionary.name == "noun"
    assert dictionary.forms == ["singular", "plural"]
    lookup = LookupObject("noun", positive_tags={"shape"})
    assert dictionary._get(lookup) == "hexagon"
    lookup.positive_tags = {"animal"}
    assert dictionary._get(lookup) == "dog"
    lookup.positive_tags = {"building", "large"}
    assert dictionary._get(lookup) == "factory"


def test_load_empty_file_raise_exception():
    path = relative_path_to_full_path("../resources/invalid_dicts/empty")
    with pytest.raises(
        TwaddleDictionaryException, match=r".*empty.dic could not be read.*"
    ):
        lookup_manager = LookupManager()
        lookup_manager.add_dictionaries_from_folder(path)
        assert len(lookup_manager.dictionaries) == 0


def test_load_file_no_forms_raise_exception():
    path = relative_path_to_full_path("../resources/invalid_dicts/no_forms")
    with pytest.raises(
        TwaddleDictionaryException, match=r".*no_forms.dic could not be read.*"
    ):
        lookup_manager = LookupManager()
        lookup_manager.add_dictionaries_from_folder(path)
        assert len(lookup_manager.dictionaries) == 0


def test_load_file_no_header_raise_exception():
    path = relative_path_to_full_path("../resources/invalid_dicts/no_header")
    with pytest.raises(
        TwaddleDictionaryException, match=r".*content_no_header.dic could not be read.*"
    ):
        lookup_manager = LookupManager()
        lookup_manager.add_dictionaries_from_folder(path)
        assert len(lookup_manager.dictionaries) == 0


def test_dictionary_manager():
    path = relative_path_to_full_path("../resources/valid_dicts")
    lookup_manager = LookupManager()
    lookup_manager.add_dictionaries_from_folder(path)
    assert len(lookup_manager.dictionaries) == 4
    noun_dictionary: LookupDictionary = lookup_manager["noun"]
    adj_dictionary: LookupDictionary = lookup_manager["adj"]
    lookup = LookupObject("noun", form="plural", positive_tags={"shape"})
    assert noun_dictionary._get(lookup) == "hexagons"
    lookup = LookupObject("adj")
    assert adj_dictionary._get(lookup) == "happy"


def test_lookup_from_object():
    path = relative_path_to_full_path("../resources/valid_dicts")
    lookup_manager = LookupManager()
    lookup_manager.add_dictionaries_from_folder(path)
    lookup = LookupObject("adj")
    assert lookup_manager.do_lookup(lookup) == "happy"


def test_lookup_indefinite_article():
    path = relative_path_to_full_path("../resources/valid_dicts")
    lookup_manager = LookupManager()
    lookup_manager.add_dictionaries_from_folder(path)
    lookup = LookupObject("article", form="indefinite")
    result = lookup_manager.do_lookup(lookup)
    assert isinstance(result, IndefiniteArticleObject)


def test_strict_lookup_invalid_tags():
    path = relative_path_to_full_path("../resources/valid_dicts")
    lookup_manager = LookupManager()
    lookup_manager.add_dictionaries_from_folder(path)
    invalid_pos_lookup = LookupObject(
        "noun", positive_tags={"invalid"}, strict_mode=True
    )
    invalid_neg_lookup = LookupObject(
        "noun", negative_tags={"invalid"}, strict_mode=True
    )
    invalid_with_valid = LookupObject(
        "noun", positive_tags={"shape", "invalid"}, strict_mode=True
    )
    invalid_with_valid_separate = LookupObject(
        "noun", positive_tags={"shape"}, negative_tags={"invalid"}, strict_mode=True
    )
    for lookup in [
        invalid_pos_lookup,
        invalid_neg_lookup,
        invalid_with_valid,
        invalid_with_valid_separate,
    ]:
        with pytest.raises(TwaddleLookupException) as e_info:
            lookup_manager.do_lookup(lookup)
        assert (
            e_info.value.message
            == "[LookupDictionary._strict_class_validation] Invalid class 'invalid'"
            " requested for dictionary 'noun' in strict mode"
        )


def test_lookup_antimatch_undefined_label():
    path = relative_path_to_full_path("../resources/valid_dicts")
    lookup_manager = LookupManager()
    lookup_manager.add_dictionaries_from_folder(path)
    strict_lookup = LookupObject(
        "noun", negative_labels={"undefined"}, strict_mode=True
    )
    with pytest.raises(TwaddleLookupException) as e_info:
        lookup_manager.do_lookup(strict_lookup)
    assert (
        e_info.value.message
        == "[LookupDictionary._strict_label_validation] Requested antimatch of label "
        "'undefined', not defined for dictionary 'noun'"
    )


def test_standard_compiler_prints_something_when_antimatch_exhausts_all_options():
    path = relative_path_to_full_path("../resources/valid_dicts")
    lookup_manager = LookupManager()
    lookup_manager.add_dictionaries_from_folder(path)
    lookup = LookupObject("noun", positive_tags={"shape"}, positive_label="a")
    assert lookup_manager.do_lookup(lookup) == "hexagon"
    lookup = LookupObject("noun", positive_tags={"shape"}, negative_labels={"a"})
    # result is random and arbitrary, simply ensure no exception is raised
    # and that _something_ is returned
    value = lookup_manager.do_lookup(lookup)
    assert value is not None


def test_strict_compiler_raises_when_antimatch_exhausts_all_options():
    path = relative_path_to_full_path("../resources/valid_dicts")
    lookup_manager = LookupManager()
    lookup_manager.add_dictionaries_from_folder(path)
    lookup = LookupObject("noun", positive_tags={"shape"}, positive_label="a")
    assert lookup_manager.do_lookup(lookup) == "hexagon"
    strict_lookup = LookupObject(
        "noun", positive_tags={"shape"}, negative_labels={"a"}, strict_mode=True
    )
    with pytest.raises(TwaddleLookupException) as e_info:
        lookup_manager.do_lookup(strict_lookup)
    assert (
        e_info.value.message
        == "[LookupDictionary._valid_choices_for_strictness_level] no valid choices for"
        " strict mode lookup in dictionary 'noun'"
    )


if __name__ == "__main__":
    test_lookup_from_object()



from collections import deque

from twaddle.lexer.lexer_tokens import Token, TokenType


def lex(input_str: str) -> deque[Token]:
    output = deque()

    i = 0
    while i < len(input_str):
        next_token, length = _get_token_type(input_str[i:])
        if next_token is not TokenType.PLAIN_TEXT:
            output.append(Token(next_token))
            i += length
        else:
            text, length = _consume_plain_text(input_str[i:])
            output.append(text)
            i += length

    return output


def _get_token_type(input_str: str) -> tuple[TokenType, int]:
    assert len(input_str) != 0
    match input_str[0]:
        case "<":
            return TokenType.LEFT_ANGLE_BRACKET, 1
        case ">":
            return TokenType.RIGHT_ANGLE_BRACKET, 1
        case "{":
            return TokenType.LEFT_CURLY_BRACKET, 1
        case "}":
            return TokenType.RIGHT_CURLY_BRACKET, 1
        case "[":
            return TokenType.LEFT_SQUARE_BRACKET, 1
        case "]":
            return TokenType.RIGHT_SQUARE_BRACKET, 1
        case "|":
            return TokenType.PIPE, 1
        case "-":
            return TokenType.HYPHEN, 1
        case ":":
            if len(input_str) > 1 and input_str[1] == ":":
                return TokenType.DOUBLE_COLON, 2
            else:
                return TokenType.COLON, 1
        case ";":
            return TokenType.SEMICOLON, 1
        case "\\":
            if len(input_str) > 1:
                match input_str[1]:
                    case "n":
                        return TokenType.NEW_LINE, 2
                    case "a":
                        return TokenType.LOWER_INDEFINITE_ARTICLE, 2
                    case "A":
                        return TokenType.UPPER_INDEFINITE_ARTICLE, 2
                    case "\\":
                        return TokenType.DOUBLE_BACKSLASH, 2
                    case "d":
                        return TokenType.DIGIT, 2
                    case "t":
                        return TokenType.TAB, 2
                    case "s":
                        return TokenType.SPACE, 2
                    case _:
                        """
                        this isn't ideal but avoids having to define special
                        tokens which only have meaning in regices
                        """
                        return TokenType.SINGLE_BACKSLASH, 1
        case '"':
            return TokenType.QUOTE, 1
        case "!":
            return TokenType.EXCLAMATION_MARK, 1
        case "=":
            return TokenType.EQUALS, 1
        case ".":
            return TokenType.DOT, 1
        case "/":
            if len(input_str) > 1 and input_str[1] == "/":
                return TokenType.REGEX, 2
            else:
                return TokenType.FORWARD_SLASH, 1
        case _:
            return TokenType.PLAIN_TEXT, 0


def _consume_plain_text(input_str: str) -> tuple[Token, int]:
    i = 0
    while i < len(input_str):
        next_type, _ = _get_token_type(input_str[i:])
        if next_type is not TokenType.PLAIN_TEXT:
            return Token(TokenType.PLAIN_TEXT, input_str[:i]), i
        else:
            i += 1

    return Token(TokenType.PLAIN_TEXT, input_str), len(input_str)



import twaddle.exceptions as rant_exceptions


def test_create_lexer_exception_with_message():
    x = rant_exceptions.TwaddleLexerException("Lexer fell over")
    assert x.message == "Lexer fell over"



from collections import OrderedDict


class DictionaryEntry:
    def __init__(self, forms: OrderedDict[str, str], tags: set[str] = None):
        self.forms = forms
        self.tags = tags

    def __getitem__(self, form: str):
        return self.forms[form]

    def has_any_tag_of(self, tags: set[str]) -> bool:
        if tags is None or self.tags is None:
            return False
        return not self.tags.isdisjoint(tags)

    def has_all_tags(self, tags: set[str]) -> bool:
        for tag in tags:
            if tag not in self.tags:
                return False
        return True



class TwaddleException(Exception):
    """Base class for others"""

    def __init__(self, message: str):
        super().__init__(message)


class TwaddleLexerException(TwaddleException):
    """Thrown when lexer gets confused"""

    def __init__(self, message: str = None):
        if message is not None:
            self.message = message
        super().__init__(self.message)


class TwaddleParserException(TwaddleException):
    """Thrown when parser gets confused"""

    def __init__(self, message: str = None):
        if message is not None:
            self.message = message
        super().__init__(self.message)


class TwaddleInterpreterException(TwaddleException):
    """Thrown when interpreter gets confused"""

    def __init__(self, message: str = None):
        if message is not None:
            self.message = message
        super().__init__(self.message)


class TwaddleLookupException(TwaddleException):
    """Thrown when lookup fails"""

    def __init__(self, message: str = None):
        if message is not None:
            self.message = message
        super().__init__(self.message)


class TwaddleDictionaryException(TwaddleException):
    """Thrown when dictionary file is invalid"""

    def __init__(self, message: str = None):
        if message is not None:
            self.message = message
        super().__init__(self.message)


class TwaddleFunctionException(TwaddleException):
    """Thrown when a function is invalid"""

    def __init__(self, message: str = None):
        if message is not None:
            self.message = message
        super().__init__(self.message)



from twaddle.interpreter.interpreter import Interpreter
from twaddle.lookup.lookup_manager import LookupManager


class TwaddleRunner:
    def __init__(
        self,
        path: str,
        persistent: bool = False,
        persistent_labels: bool = False,
        persistent_synchronizers: bool = False,
        persistent_patterns: bool = False,
        strict_mode: bool = False,
    ):
        self.lookup_manager = LookupManager()
        self.lookup_manager.add_dictionaries_from_folder(path)
        persistent_labels = True if persistent else persistent_labels
        persistent_synchronizers = True if persistent else persistent_synchronizers
        persistent_patterns = True if persistent else persistent_patterns
        self.interpreter = Interpreter(
            self.lookup_manager,
            persistent_labels=persistent_labels,
            persistent_synchronizers=persistent_synchronizers,
            persistent_patterns=persistent_patterns,
            strict_mode=strict_mode,
        )
        pass

    def run_sentence(self, sentence: str) -> str:
        return self.interpreter.interpret_external(sentence)

    def clear(self) -> None:
        self.interpreter.force_clear()



import pytest

from twaddle.exceptions import TwaddleInterpreterException
from twaddle.interpreter.interpreter import Interpreter
from twaddle.lookup.lookup_manager import LookupManager

standard_interpreter = Interpreter(LookupManager())


def get_standard_interpreter_output(sentence: str) -> str:
    return standard_interpreter.interpret_external(sentence)


def test_paste_copyd_block():
    result = get_standard_interpreter_output("[copy:word]{word} [paste:word]")
    assert result == "word word"


def test_paste_uncopyd_block():
    with pytest.raises(TwaddleInterpreterException) as e_info:
        get_standard_interpreter_output("[paste:nonexistent]")
    assert (
        e_info.value.message
        == "[Interpreter._handle_special_functions#paste] Tried to paste result of unknown block 'nonexistent'"
    )


def test_copy_multiple_blocks():
    result = get_standard_interpreter_output(
        "[copy:first]{one} [copy:second]{two} [paste:first] [paste:second]"
    )
    assert result == "one two one two"


def test_copy_paste_fails_between_non_persistent_sentences():
    get_standard_interpreter_output("[copy:a]{hello}")
    with pytest.raises(TwaddleInterpreterException) as e_info:
        get_standard_interpreter_output("[paste:a]")
    assert (
        e_info.value.message
        == "[Interpreter._handle_special_functions#paste] Tried to paste result of unknown block 'a'"
    )


def test_copy_paste_persistence():
    interpreter = Interpreter(LookupManager(), persistent_clipboard=True)
    interpreter.interpret_external("[copy:a]{hello}")
    assert interpreter.interpret_external("[paste:a]") == "hello"



from twaddle.interpreter.synchronizer import (
    CyclicDeckSynchronizer,
    DeckSynchronizer,
    LockedSynchronizer,
    SynchronizerManager,
)


def test_synchronizer_manager():
    sync_manager = SynchronizerManager()
    locked = sync_manager.create_synchronizer("x", "locked", 1)
    assert isinstance(locked, LockedSynchronizer)


def test_locked_synchronizer():
    locked = LockedSynchronizer(10)
    pick = locked.next()
    for _ in range(0, 10):
        assert locked.next() == pick


def test_deck_synchronizer():
    deck = DeckSynchronizer(10)
    results = dict[int:int]()
    for value in range(0, 10):
        results[value] = 0
    for _ in range(0, 20):
        value = deck.next()
        results[value] = results[value] + 1
    for value in range(0, 10):
        assert results[value] == 2


# noinspection SpellCheckingInspection
def test_cyclic_deck_synchronizer():
    cdeck = CyclicDeckSynchronizer(10)
    results1 = list()
    results2 = list()

    for _ in range(0, 10):
        results1.append(cdeck.next())
    for _ in range(0, 10):
        results2.append(cdeck.next())
    for n in range(0, 10):
        assert results1[n] == results2[n]



import readline  # noqa: F401
import sys
from importlib.resources import files

from twaddle.exceptions import TwaddleException
from twaddle.runner import TwaddleRunner


def main():
    if len(sys.argv) < 2:
        path = files("twaddle.default_dictionary")
    else:
        path = sys.argv[1]
    twaddle = TwaddleRunner(path)

    print("hello. I'm your friendly nonsense generator. Hit Ctrl-D to exit.")

    while True:
        try:
            sentence = input(">")
            print(twaddle.run_sentence(sentence))
        except TwaddleException as te:
            print(f"Twaddle encountered an error:\n{te.message}")
            twaddle.clear()
        except EOFError:
            quit()


if __name__ == "__main__":
    main()



from random import randint

from twaddle.compiler.compiler_objects import RootObject
from twaddle.exceptions import TwaddleFunctionException
from twaddle.interpreter.block_attributes import BlockAttributeManager
from twaddle.interpreter.formatting_object import FormattingStrategy
from twaddle.interpreter.regex_state import RegexState


def repeat(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    repetitions = int(evaluated_args[0])
    block_attribute_manager.current_attributes.repetitions = repetitions


def separator(
    _evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    raw_args: list[RootObject],
):
    block_attribute_manager.current_attributes.separator = raw_args[0]


def first(
    _evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    raw_args: list[RootObject],
):
    block_attribute_manager.current_attributes.first = raw_args[0]


def last(
    _evaluated_args,
    block_attribute_manager: BlockAttributeManager,
    raw_args: list[RootObject],
):
    block_attribute_manager.current_attributes.last = raw_args[0]


def save(
    evaluated_args: list[str], block_attribute_manager: BlockAttributeManager, _raw_args
):
    block_attribute_manager.save_block(evaluated_args[0])


def copy(
    evaluated_args: list[str], block_attribute_manager: BlockAttributeManager, _raw_args
):
    block_attribute_manager.copy_block(evaluated_args[0])


def sync(
    evaluated_args: list[str], block_attribute_manager: BlockAttributeManager, _raw_args
):
    block_attribute_manager.set_synchronizer(evaluated_args)


def abbreviate(
    evaluated_args: list[str], block_attribute_manager: BlockAttributeManager, _raw_args
):
    block_attribute_manager.current_attributes.abbreviate = True
    if len(evaluated_args) == 0:
        block_attribute_manager.current_attributes.abbreviation_case = (
            FormattingStrategy.UPPER
        )
        return
    case = evaluated_args[0].strip().lower()
    match case:
        case "retain":
            block_attribute_manager.current_attributes.abbreviation_case = (
                FormattingStrategy.NONE
            )
        case "upper":
            block_attribute_manager.current_attributes.abbreviation_case = (
                FormattingStrategy.UPPER
            )
        case "lower":
            block_attribute_manager.current_attributes.abbreviation_case = (
                FormattingStrategy.LOWER
            )
        case "first":
            block_attribute_manager.current_attributes.abbreviation_case = (
                FormattingStrategy.TITLE
            )
        case _:
            raise TwaddleFunctionException(
                "[function_definitions#abbreviate] invalid case " f"argument '{case}'"
            )


def case(evaluated_args: list[str], _block_attribute_manager, _raw_args):
    arg = evaluated_args[0].strip().lower()
    match arg:
        case "none":
            return FormattingStrategy.NONE
        case "upper":
            return FormattingStrategy.UPPER
        case "lower":
            return FormattingStrategy.LOWER
        case "sentence":
            return FormattingStrategy.SENTENCE
        case "title":
            return FormattingStrategy.TITLE
        case _:
            pass


# noinspection PyUnusedLocal
def match(evaluated_args: list[str], _block_attribute_manager, _raw_args):
    return RegexState.match


def rand(evaluated_args: list[str], _block_attribute_manager, _raw_args) -> str:
    minimum = int(evaluated_args[0])
    maximum = int(evaluated_args[1])
    return str(randint(minimum, maximum))


def reverse(_evaluated_args: list[str], block_attribute_manager, _raw_args):
    block_attribute_manager.current_attributes.reverse = True


def hide(
    evaluated_args: list[str], block_attribute_manager: BlockAttributeManager, _raw_args
) -> str:
    block_attribute_manager.current_attributes.hidden = True



from twaddle.interpreter.block_attributes import BlockAttributeManager, BlockAttributes


def test_add_and_get_attributes():
    bam = BlockAttributeManager()
    bam.current_attributes.repetitions = 3
    bam.current_attributes.separator = "and"
    attributes: BlockAttributes = bam.get_attributes()
    new_attributes: BlockAttributes = bam.get_attributes()
    assert attributes.repetitions == 3
    assert attributes.separator == "and"
    assert new_attributes.repetitions == 1
    assert new_attributes.separator is None



from pathlib import Path

from twaddle.compiler.compiler_objects import LookupObject
from twaddle.exceptions import TwaddleDictionaryException
from twaddle.lookup.dictionary_file_parser import DictionaryFileParser
from twaddle.lookup.lookup_dictionary import LookupDictionary


class LookupManager:
    def __init__(self):
        self.dictionaries = dict[str, LookupDictionary]()

    def __getitem__(self, name: str) -> LookupDictionary:
        if dictionary := self.dictionaries.get(name):
            return dictionary
        raise TwaddleDictionaryException(
            f"[LookupManager.__getitem__] No dictionary loaded named {name}"
        )

    def add_dictionaries_from_folder(self, folder: str | Path):
        if isinstance(folder, (str, Path)):
            folder = Path(folder)

        for entry in folder.iterdir():
            if entry.name.endswith(".dic") and entry.is_file():
                new_dictionary = DictionaryFileParser.read_from_path(entry)
                if new_dictionary is None:
                    raise TwaddleDictionaryException(
                        f"[LookupManager.add_dictionaries_from_folder] dictionary file {entry} could not be read."
                        "Are name and forms defined?"
                    )
                self.dictionaries[new_dictionary.name] = new_dictionary

    def clear_labels(self):
        for dictionary in self.dictionaries.values():
            dictionary.clear_labels()

    def do_lookup(self, lookup: LookupObject):
        dictionary: LookupDictionary = self.dictionaries[lookup.dictionary]
        return dictionary.get(lookup)



from enum import Enum, auto
from typing import Type


class FormattingStrategy(Enum):
    NONE = auto()
    UPPER = auto()
    LOWER = auto()
    SENTENCE = auto()
    TITLE = auto()


class FormattingObject:
    def __init__(self, previous):
        self.previous = None
        if previous is not None:
            self.previous = previous
            previous.next = self
        self.next = None


class PlainText(FormattingObject):
    def __init__(self, previous: Type[FormattingObject], text: str):
        FormattingObject.__init__(self, previous)
        self.text = text


class StrategyChange(FormattingObject):
    def __init__(self, previous: Type[FormattingObject], strategy: FormattingStrategy):
        FormattingObject.__init__(self, previous)
        self.strategy = strategy


class IndefiniteArticle(FormattingObject):
    def __init__(
        self, previous: Type[FormattingObject], default_upper_case: bool = False
    ):
        FormattingObject.__init__(self, previous)
        self.default_upper = default_upper_case



from abc import ABC, abstractmethod
from random import randrange, shuffle

from twaddle.exceptions import TwaddleInterpreterException


class Synchronizer(ABC):
    def __init__(self, num_choices: int):
        self.num_choices = num_choices

    @abstractmethod
    def next(self) -> int:
        pass


class LockedSynchronizer(Synchronizer):
    def __init__(self, num_choices: int):
        super().__init__(num_choices)
        self.pick = randrange(0, num_choices)

    def next(self) -> int:
        return self.pick


class DeckSynchronizer(Synchronizer):
    def __init__(self, num_choices: int):
        super().__init__(num_choices)
        self.num_choices = num_choices
        self.deck = list()
        self.shuffle_deck()

    def shuffle_deck(self):
        self.deck = list(range(0, self.num_choices))
        shuffle(self.deck)

    def next(self) -> int:
        if len(self.deck) == 0:
            self.shuffle_deck()
        return self.deck.pop(0)


class CyclicDeckSynchronizer(Synchronizer):
    def __init__(self, num_choices: int):
        super().__init__(num_choices)
        self.num_choices = num_choices
        self.deck = list()
        self.shuffle_deck()
        self.pos = 0

    def shuffle_deck(self):
        self.deck = list(range(0, self.num_choices))
        shuffle(self.deck)

    def next(self) -> int:
        self.pos = self.pos + 1
        if self.pos == self.num_choices:
            self.pos = 0
        return self.deck[self.pos]


class SynchronizerManager:

    sync_types = {
        "locked": LockedSynchronizer,
        "deck": DeckSynchronizer,
        "cdeck": CyclicDeckSynchronizer,
    }

    def __init__(self):
        self.synchronizers = dict[str, Synchronizer]()

    def synchronizer_exists(self, name: str) -> bool:
        return name in self.synchronizers

    def create_synchronizer(
        self, name: str, sync_type: str, length: int
    ) -> Synchronizer:
        self.synchronizers[name] = SynchronizerManager.sync_types[sync_type](length)
        return self.synchronizers[name]

    def get_synchronizer(self, name: str) -> Synchronizer:
        if not self.synchronizer_exists(name):
            raise TwaddleInterpreterException(
                f"[SynchronizerManager.get_synchronizer] tried to access non-existing synchronizer {name}"
            )
        return self.synchronizers[name]

    def clear(self):
        self.synchronizers.clear()



from twaddle.lexer.lexer_tokens import Token, TokenType


def test_rant_token_equality():
    plaintext = Token(TokenType.PLAIN_TEXT, "plaintext")
    duplicate_plaintext = Token(TokenType.PLAIN_TEXT, "plaintext")
    different_plaintext = Token(TokenType.PLAIN_TEXT, "different plaintext")
    empty_plaintext = Token(TokenType.PLAIN_TEXT)
    left_angle = Token(TokenType.LEFT_ANGLE_BRACKET)
    hyphen = Token(TokenType.HYPHEN)

    assert plaintext == duplicate_plaintext
    assert plaintext != empty_plaintext
    assert plaintext != left_angle
    assert plaintext != different_plaintext
    assert left_angle != hyphen
